#!/usr/bin/env python

"""
Unit and integration tests for allele_tools/stec.py allele_concatenate
"""

# Standard imports
from unittest.mock import patch
import argparse
import shutil
import os

# Third-party imports
import pytest

# Local imports
from allele_tools.allele_profiler import read_profile
from allele_tools.methods import (
    concatenate_alleles,
    load_alleles,
    write_concatenated_sequences
)
from allele_tools.stec import (
    allele_concatenate,
    cli
)


@pytest.fixture(name='variables', scope='module')
def setup():
    class Variables:
        def __init__(self):
            # Extract the connection string prior to running tests
            self.test_path = os.path.abspath(os.path.dirname(__file__))
            self.file_path = os.path.join(self.test_path, 'test_files', 'allele_updater', 'original_files')
            self.nt_allele_path = os.path.join(self.file_path, 'nt_alleles')
            self.aa_allele_path = os.path.join(self.file_path, 'aa_alleles')
            self.nt_profile_file = os.path.join(self.file_path, 'nt_profile', 'profile.txt')
            self.aa_profile_file = os.path.join(self.file_path, 'aa_profile', 'profile.txt')
            self.concatenate_path = os.path.join(self.test_path, 'test_files', 'concatenate_alleles')
            self.linker_length_dict = {
                'stx1': 9,
                'stx2': 12,
            }
            self.allele_order = {
                'stx1': ['ECs2974', 'ECs2973'],
                'stx2': ['ECs1205', 'ECs1206']
            }
            self.gene_allele = {
                'stx1': 'ECs2974_ECs2973',
                'stx2': 'ECs1205_ECs1206'
            }

    return Variables()


def prepare_files(variables):
    if not os.path.isdir(variables.concatenate_path):
        os.makedirs(variables.concatenate_path)
    assert os.path.isdir(variables.concatenate_path)


def clear_path(variables):
    shutil.rmtree(variables.concatenate_path)
    assert not os.path.isdir(variables.concatenate_path)


def test_read_profile(variables):
    variables.nt_profile_data = read_profile(
        profile_file=variables.nt_profile_file
    )
    assert variables.nt_profile_data['113'] == {
        'ECs1205': '12',
        'ECs1206': '1'
    }


def test_load_alleles(variables):
    variables.gene, variables.nt_alleles = load_alleles(
        allele_path=variables.nt_allele_path,
        allele_order=variables.allele_order
    )
    assert variables.gene == 'stx2'
    assert variables.nt_alleles['ECs1205']['ECs1205_1'][:50] == 'ATGAAGTGTATATTATTTAAATGGGTACTGTGCCTGTTACTGGGTTTTTC'


def test_concatenate_alleles(variables):
    variables.concatenated_nt_seq = concatenate_alleles(
        profile_data=variables.nt_profile_data,
        allele_dict=variables.nt_alleles,
        allele_order=variables.allele_order,
        stx_gene=variables.gene,
        linker_length_dict=variables.linker_length_dict,
        molecule='nt'
    )
    assert str(variables.concatenated_nt_seq[0].seq)[:50] == 'ATGAAGTGTATATTATTTAAATGGGTACTGTGCCTGTTACTGGGTTTTTC'
    assert str(variables.concatenated_nt_seq[0].seq)[-50:] == 'CCTGTGAATCAGGCTCCGGATTTGCTGAAGTGCAGTTTAATAATGACTGA'


def test_write_concatenated_alleles(variables):
    prepare_files(variables=variables)
    write_concatenated_sequences(
        concatenated_sequences=variables.concatenated_nt_seq,
        concatenate_path=variables.concatenate_path,
        file_name=variables.gene_allele[variables.gene],
        molecule='nt'
    )
    assert os.path.isfile(os.path.join(variables.concatenate_path, 'nt', 'ECs1205_ECs1206.fasta'))
    clear_path(variables=variables)


@patch('argparse.ArgumentParser.parse_args')
def test_allele_concatenate_integration(mock_args, variables):
    prepare_files(variables=variables)
    mock_args.return_value = argparse.Namespace(
        nt_profile=variables.nt_profile_file,
        aa_profile=variables.aa_profile_file,
        nt_alleles=variables.nt_allele_path,
        aa_alleles=variables.aa_allele_path,
        concatenate_path=variables.concatenate_path
    )
    arguments = cli()
    allele_concatenate(args=arguments)
    assert os.path.isfile(os.path.join(variables.concatenate_path, 'aa', 'ECs1205_ECs1206.fasta'))
    clear_path(variables=variables)



#!/usr/bin/env python

"""
Unit and integration tests for allele_tools/profile_reduce.py
"""

# Standard imports
from unittest.mock import patch
import argparse
import shutil
import os

# Third-party imports
import pytest

# Local imports
from allele_tools.profile_reduce import \
    cli


@pytest.fixture(name='variables', scope='module')
def setup():
    class Variables:
        def __init__(self):
            # Extract the connection string prior to running tests
            self.test_path = os.path.abspath(os.path.dirname(__file__))
            self.file_path = os.path.join(self.test_path, 'test_files', 'profile_reduce')
            self.profile = os.path.join(self.file_path, 'profiles.list')
            self.names = os.path.join(self.file_path, 'genes.txt')
            self.output_path = os.path.join(self.file_path, 'profile')
            self.output_profiles = os.path.join(self.file_path, 'profile', 'profile.txt')
            self.output_notes = os.path.join(self.file_path, 'profile', 'reducing_notes.txt')

    return Variables()


@patch('argparse.ArgumentParser.parse_args')
def test_profile_reduce_missing_tilde_profile(mock_args, variables):
    mock_args.return_value = argparse.Namespace(
        profile='~/fake_file_profiles.list',
        names=variables.names
    )
    with pytest.raises(SystemExit):
        cli()


@patch('argparse.ArgumentParser.parse_args')
def test_profile_reduce_missing_tilde_names(mock_args, variables):
    mock_args.return_value = argparse.Namespace(
        profile=variables.profile,
        names='~/fake_file_genes.txt'
    )
    with pytest.raises(SystemExit):
        cli()


@patch('argparse.ArgumentParser.parse_args')
def test_profile_reduce_integration(mock_args, variables):
    mock_args.return_value = argparse.Namespace(
        profile=variables.profile,
        names=variables.names
    )
    cli()
    assert os.path.isfile(variables.output_profiles)


def test_output_profile_header(variables):
    variables.profiles = open(variables.output_profiles, 'r', encoding='utf-8').readlines()
    assert variables.profiles[0] == 'ST	ECs1205	ECs1206\n'


def test_output_profile_sequence_type(variables):
    assert variables.profiles[1] == '1	0	0\n'


def test_output_notes_header(variables):
    variables.notes = open(variables.output_notes, 'r', encoding='utf-8').readlines()
    assert variables.notes[0] == 'OriginalSequenceType	ReducedSequenceType	Notes\n'


def test_output_notes_novel(variables):
    assert variables.notes[1] == '1	1\n'


def test_output_notes_duplicate(variables):
    assert variables.notes[2] == '2	0	duplicate\n'


def test_output_notes_length(variables):
    assert len(variables.notes) == 10


def test_output_clean_up(variables):
    shutil.rmtree(variables.output_path)
    assert not os.path.isfile(variables.output_path)



#!/usr/bin/env python

"""
Unit and integration tests for allele_tools/stec.py Only testing the profile_reduce functionality
"""

# Standard imports
from unittest.mock import patch
import argparse
import shutil
import os

import pytest

# Local imports
from allele_tools.stec import \
    translate_reduce, \
    cli
from .test_1_allele_translate_reduce import \
    setup, \
    clean_outputs

assert vars(setup)['_pytestfixturefunction'].name == 'variables'


def prepare_files(variables):
    if not os.path.isdir(variables.nt_profile_path):
        os.makedirs(variables.nt_profile_path)
    shutil.copyfile(
        src=variables.nt_profile_file,
        dst=os.path.join(variables.nt_profile_path, 'profile.txt')
    )
    if not os.path.isdir(variables.nt_allele_path):
        os.makedirs(variables.nt_allele_path)
    for allele_file in variables.nt_allele_files:
        file_name = os.path.basename(allele_file)
        shutil.copyfile(
            src=allele_file,
            dst=os.path.join(variables.nt_allele_path, file_name)
        )


@patch('argparse.ArgumentParser.parse_args')
def test_allele_translate_reduce_integration(mock_args, variables):
    prepare_files(variables=variables)
    mock_args.return_value = argparse.Namespace(
        allele_path=variables.nt_allele_path,
        profile_file=variables.nt_profile,
        report_path=variables.aa_profile_path,
        translated_path=variables.aa_allele_path
    )
    arguments = cli()
    translate_reduce(args=arguments)


def test_stec_profile_header(variables):
    from .test_1_allele_translate_reduce import test_profile_header
    test_profile_header(variables=variables)


def test_stec_profile_contents(variables):
    from .test_1_allele_translate_reduce import test_profile_contents
    test_profile_contents(variables=variables)


def test_stec_profile_end(variables):
    from .test_1_allele_translate_reduce import test_profile_end
    test_profile_end(variables=variables)


def test_stec_clean_outputs(variables):
    clean_outputs(variables=variables)


@patch('argparse.ArgumentParser.parse_args')
def test_allele_translate_reduce_integration_no_gene_path(mock_args, variables):
    with pytest.raises(SystemExit):
        mock_args.return_value = argparse.Namespace(
            allele_path=variables.nt_allele_path,
            profile_file=variables.nt_profile,
            report_path=variables.aa_profile_path,
            translated_path=variables.aa_allele_path
        )
        arguments = cli()
        translate_reduce(args=arguments)


@patch('argparse.ArgumentParser.parse_args')
def test_allele_translate_reduce_integration_no_gene_file(mock_args, variables):
    with pytest.raises(SystemExit):
        mock_args.return_value = argparse.Namespace(
            allele_path=variables.nt_allele_path,
            profile_file=variables.nt_profile,
            report_path=variables.aa_profile_path,
            translated_path=variables.aa_allele_path
        )
        arguments = cli()
        translate_reduce(args=arguments)



#!/usr/bin/env python

"""
Determine allele complement and profile of sequences. Update profiles and databases as necessary
"""

# Standard imports
from argparse import ArgumentParser
import multiprocessing
from glob import glob
import logging
import json
import os


# Third-party imports
from olctools.accessoryFunctions.accessoryFunctions import \
    make_path, \
    MetadataObject, \
    SetupLogging
from Bio import SeqIO

# Local imports
from allele_tools.allele_profiler import \
    allele_prep, \
    append_profiles, \
    clear_alleles, \
    create_profile, \
    match_profile, \
    parse_results, \
    parseable_blast_outputs, \
    profile_alleles, \
    read_profile, \
    sequence_typer
from allele_tools.methods import \
    update_allele_database, \
    translate, \
    query_prep, \
    blast_alleles, \
    pathfinder

__author__ = 'adamkoziol'


class Updater:

    """
    Determine allele complement and profile of sequences. Update profiles and databases as
    necessary
    """

    # pylint: disable=too-many-instance-attributes

    def main(self):
        """
        Run the appropriate methods in the correct order
        """
        # Create metadata objects for all files in the query folder
        self.runmetadata = query_prep(
            query_path=self.query_path,
            runmetadata=self.runmetadata
        )
        for sample in self.runmetadata.samples:
            logging.debug('Processing sample %s', sample.name)
            # Perform necessary prep on the alleles
            if not self.amino_acid:
                records, gene_names, self.data = \
                    allele_prep(
                        allele_path=self.allele_path,
                        gene_names=self.gene_names,
                        combined_targets=self.combined_targets,
                        amino_acid=self.amino_acid
                    )
            else:
                records, gene_names, self.data = \
                    allele_prep(
                        allele_path=self.aa_allele_path,
                        gene_names=self.gene_names,
                        combined_targets=self.combined_targets,
                        amino_acid=self.amino_acid
                    )

            logging.info('Loading profile')
            if not self.amino_acid:
                profile_data = read_profile(profile_file=self.profile_file)
            else:
                profile_data = read_profile(profile_file=self.aa_profile_file)
            # BLAST the query against the allele database
            blast_alleles(
                runmetadata=sample,
                amino_acid=self.amino_acid,
                combined_targets=self.combined_targets,
                cpus=self.cpus,
                outfmt=self.outfmt
            )
            # Add a user-friendly header to the BLAST outputs
            parseable_blast_outputs(
                runmetadata=sample,
                fieldnames=self.fieldnames,
                extended_fieldnames=self.extended_fieldnames,
                records=records
            )
            # Parse the BLAST outputs
            sample = parse_results(
                runmetadata=sample,
                fieldnames=self.fieldnames,
                extended_fieldnames=self.extended_fieldnames,
                amino_acid=self.amino_acid,
                genome_query=True
            )
            # Perform sequence typing of the parsed results
            if not self.amino_acid:
                profile_dict, profile_set = profile_alleles(
                    runmetadata=sample,
                    profile_dict={},
                    profile_set=[],
                    records=self.gene_names,
                    novel_alleles=True,
                    genome_query=True,
                    allele_path=self.allele_path,
                    report_path=self.report_path
                )
            else:
                profile_dict, profile_set = profile_alleles(
                    runmetadata=sample,
                    profile_dict={},
                    profile_set=[],
                    records=self.gene_names,
                    amino_acid=True,
                    novel_alleles=True,
                    genome_query=True,
                    allele_path=self.aa_allele_path,
                    report_path=self.report_path
                )
            # Match the query profile against the profile database
            profile_matches = match_profile(
                profile_data=profile_data,
                profile_dict=profile_dict,
                profile_matches={}
            )
            # Create new profiles as required
            profile_matches, profile_data, new_profiles = \
                create_profile(
                    profile_data=profile_data,
                    profile_set=profile_set,
                    new_profiles=[],
                    profile_dict=profile_dict,
                    profile_matches=profile_matches
                )
            # Perform final sequence typing, and create final report
            if not self.amino_acid:
                sample = sequence_typer(
                    profile_report=self.profile_report,
                    data=self.data,
                    runmetadata=sample,
                    profile_matches=profile_matches,
                    profile_data=profile_data,
                    update=True
                )
                # Write the novel profiles to file
                append_profiles(
                    new_profiles=new_profiles,
                    profile_file=self.profile_file,
                    data=self.data,
                    novel_profiles=True,
                    profile_path=self.profile_path,
                    gene_names=self.gene_names
                )
            else:
                sample = sequence_typer(
                    profile_report=self.aa_profile_report,
                    data=self.data,
                    runmetadata=sample,
                    profile_matches=profile_matches,
                    profile_data=profile_data,
                    update=True
                )
                append_profiles(
                    new_profiles=new_profiles,
                    profile_file=self.aa_profile_file,
                    data=self.data,
                    novel_profiles=True,
                    profile_path=self.aa_profile_path,
                    gene_names=self.gene_names
                )
            if not self.amino_acid:
                # AA
                sample = translate(runmetadata=sample)
                self.aa_allele_prep()
                aa_profile_dict, aa_profile_set = self.aa_allele_match(
                    runmetadata=sample,
                    profile_dict={},
                    profile_set=[],
                    gene_names=gene_names,
                    amino_acid=True
                )
                aa_profile_data = read_profile(profile_file=self.aa_profile_file)
                aa_profile_matches = match_profile(
                    profile_data=aa_profile_data,
                    profile_dict=aa_profile_dict,
                    profile_matches={}
                )
                aa_profile_matches, aa_profile_data, aa_new_profiles = \
                    create_profile(
                        profile_data=aa_profile_data,
                        profile_set=aa_profile_set,
                        new_profiles=[],
                        profile_dict=aa_profile_dict,
                        profile_matches=aa_profile_matches
                    )

                sample = sequence_typer(
                    profile_report=self.aa_profile_report,
                    data=self.data,
                    runmetadata=sample,
                    profile_matches=aa_profile_matches,
                    profile_data=aa_profile_data,
                    update=True,
                    amino_acid=True
                    )
                make_path(self.aa_profile_path)
                append_profiles(
                    new_profiles=aa_new_profiles,
                    profile_file=self.aa_profile_file,
                    data=self.data,
                    novel_profiles=True,
                    profile_path=self.aa_profile_path,
                    gene_names=self.gene_names
                )
                self.aa_notes(runmetadata=sample)
                clear_alleles(
                    combined_targets_db=glob(os.path.join(self.allele_path, 'combinedtargets*')),
                    custom_targets=os.path.join(self.allele_path, 'custom.tfa')
                )

    def aa_allele_prep(self):
        """
        Create (first time only) and read the amino acid allele database file
        """
        # Create the amino acid allele database file path as required
        make_path(self.aa_allele_path)
        # Iterate through all the gene in the analysis
        for gene in self.gene_names:
            # Attempt to find the database file
            try:
                allele_file = glob(os.path.join(self.aa_allele_path, f'{gene}*.*fa*'))[0]
            # Create the file if it doesn't exist
            except IndexError:
                allele_file = self.initialise_aa_alleles(gene=gene)
            # Read in and store all the amino acid records in the allele database file
            for record in SeqIO.parse(allele_file, 'fasta'):
                self.aa_allele_dict[record.id] = str(record.seq)

    def initialise_aa_alleles(self, gene):
        """
        Create a gene-specific amino acid allele database file
        :param gene: Name of current gene being analysed
        :return: Name and absolute path to the created database file
        """
        # Find the corresponding gene-specific nucleotide database file
        nt_allele_file = glob(os.path.join(self.allele_path, f'{gene}*.*fa*'))[0]
        # Set the name of the amino acid database file
        aa_allele_file = os.path.join(self.aa_allele_path, f'{gene}_alleles.fasta')
        # Initialise a dictionary to store str(amino acid sequence): allele name to be used
        # in finding duplicate
        # translated alleles
        allele_dict = {}
        for record in SeqIO.parse(nt_allele_file, 'fasta'):
            # Replace and dashes in the record.id with underscores
            record.id = record.id.replace('-', '_')
            record_id = record.id
            # Translate the sequence to amino acid
            record = record.translate()
            record.id = record_id
            record.description = str()
            # Extract the gene name from the allele number
            gene, allele_id = record.id.rsplit('_', 1)
            # Initialise the gene key in the dictionary as required
            if gene not in self.aa_nt_allele_link_dict:
                self.aa_nt_allele_link_dict[gene] = dict()
            # Check if the translated sequence is not present in the dictionary
            if str(record.seq) not in allele_dict:
                # Update the dictionary with the sequence: allele
                allele_dict[str(record.seq)] = record.id
                # Update the dictionary with gene: allele
                if not self.aa_nt_allele_link_dict[gene]:
                    self.aa_nt_allele_link_dict[gene][record.id] = record.id
                if allele_id == '1':
                    # Write the translated target sequence to file
                    with open(aa_allele_file, 'a+', encoding='utf-8') as aa_targets:
                        SeqIO.write(record, aa_targets, 'fasta')
            else:
                if not self.aa_nt_allele_link_dict[gene]:
                    self.aa_nt_allele_link_dict[gene][record.id] = allele_dict[str(record.seq)]
        return aa_allele_file

    def aa_allele_match(self, runmetadata, profile_dict, profile_set, gene_names, amino_acid):
        """
        Find match the alleles in the current sample to alleles in the database
        :param runmetadata: List of metadata objects for the current sample
        :param profile_dict: Dictionary to store gene:allele profile for each sample
        :param profile_set: List of all unique profiles
        :param gene_names: List of all gene names in the analysis
        :param amino_acid: Variable indicating whether the current analyses are on DNA or
        amino acid sequences
        :return: profile_dict and profile_set updated with the results from the current sample
        """
        for sample in runmetadata.samples:
            # Initialise a dictionary to store the gene:allele combinations for each sample
            allele_comprehension = {}
            for allele, allele_seq in sorted(sample.alleles.nt_alleles_translated.items()):
                present = False
                # Strip off the allele number from the allele e.g. adk_1 yields adk, while EC042_RS26480_2 yields
                # EC042_RS26480
                try:
                    gene, _ = allele.rsplit('_', 1)
                except ValueError:
                    gene = str()
                for record_id, record_seq in sorted(self.aa_allele_dict.items()):
                    if allele_seq == record_seq:
                        # Update the dictionary with the new gene: allele number for the sample
                        allele_comprehension.update({gene: record_id.split('_')[-1]})
                        present = True
                # If, after iterating through all the BLAST outputs, the gene is not present in the sample, update
                # the gene: allele to reflect this absence
                if not present:
                    novel_allele = update_allele_database(
                        gene=gene,
                        query_sequence=allele_seq,
                        allele_path=self.aa_allele_path,
                        report_path=self.report_path,
                        amino_acid=amino_acid
                    )
                    if novel_allele:
                        allele_comprehension.update({gene: novel_allele.split('_')[-1]})
                    else:
                        allele_comprehension.update({gene: '0'})
            profile_dict, profile_set, allele_comprehension = self.profile_dict_set_populate(
                gene_names=gene_names,
                allele_comprehension=allele_comprehension,
                profile_dict=profile_dict,
                sample=sample,
                profile_set=profile_set
            )
        return profile_dict, profile_set

    @staticmethod
    def profile_dict_set_populate(gene_names, allele_comprehension, profile_dict, sample, profile_set):
        """
        Update the profile dictionary, profile set, and allele comprehension for each gene
        :param gene_names: List of all gene names in the analysis
        :param allele_comprehension: Dictionary of the gene:allele combinations for each sample
        :param profile_dict: Dictionary to store gene:allele profile for each sample
        :param sample: Metadata objects of current genome
        :param profile_set: List of all unique profiles
        """
        # Iterate through all the genes
        for gene_name in gene_names:
            if gene_name not in allele_comprehension:
                allele_comprehension.update({gene_name: '0'})
            # In order to hash the dictionary, use JSON, with sorted keys to freeze it
            frozen_allele_comprehension = json.dumps(allele_comprehension, sort_keys=True)
            # Update the dictionary of profiles with the hash of the
            # frozen dictionary: list of samples with that hash
            if hash(frozen_allele_comprehension) not in profile_dict:
                profile_dict[hash(frozen_allele_comprehension)] = [sample.name]
            else:
                profile_dict[hash(frozen_allele_comprehension)].append(sample.name)
            # Add the 'regular' dictionary to the list of all profiles as required
            if allele_comprehension not in profile_set:
                profile_set.append(allele_comprehension)
        return profile_dict, profile_set, allele_comprehension

    def aa_notes(self, runmetadata):
        """
        Create (first time only), and update the profile and alleles notes files. These files
        link amino acid profile(s) and allele(s), respectively, to the corresponding nucleotide
        profile, and allele
        :param runmetadata: List of metadata objects for the current sample
        """
        logging.info('Creating/Updating notes')
        # Read in the profile notes file
        allele_profile_dict = self.read_aa_profile_notes()
        for sample in runmetadata.samples:
            # Check if the nucleotide sequence type has previously been encountered
            try:
                # Extract all the previously encountered amino acid sequence types corresponding
                # to the nucleotide st
                known_allele_profiles = allele_profile_dict[sample.alleles.nt_st]
                # Determine whether the current amino acid st is already in the list of aa
                # sequence types
                if sample.alleles.aa_st not in known_allele_profiles:
                    # If the current st is novel, update the list of profiles, and use the
                    # list to update the
                    # notes file
                    allele_profile_dict[sample.alleles.nt_st].append(sample.alleles.aa_st)
                    self.update_aa_profile_notes(aa_profile_dict=allele_profile_dict)
            # If the current nucleotide sequence type is novel, add it to the dictionary
            except KeyError:
                allele_profile_dict[sample.alleles.nt_st] = [sample.alleles.aa_st]
                # Use the dictionary to update the notes file
                self.update_aa_profile_notes(aa_profile_dict=allele_profile_dict)
            # Process the allele file
            for gene, allele in sample.alleles.nt_profile.items():
                # Attempt to read in the previous allele notes file
                allele_notes_dict = self.read_aa_allele_notes(gene=gene)
                # Check if the nucleotide allele is present in the dictionary
                try:
                    # Extract the list of all amino acid alleles corresponding
                    # to the nucleotide allele
                    try:
                        known_aa_alleles = allele_notes_dict[int(allele)]
                    # Initialise an empty list if allele_notes_dict is empty
                    except ValueError:
                        known_aa_alleles = list()
                    # Set the aa_allele variable to save typing the long attribute
                    aa_allele = sample.alleles.aa_profile[gene]
                    # Check to see if the amino acid allele has already been encountered
                    if aa_allele not in known_aa_alleles:
                        # Make sure that the allele isn't missing 'ND'
                        if aa_allele != '0':
                            # Add the allele to the list
                            allele_notes_dict[int(allele)].append(aa_allele)
                            # Update the notes file with the
                            self.update_aa_allele_notes(
                                gene=gene,
                                allele_notes_dict=allele_notes_dict
                            )
                # If the nucleotide allele is novel, add it to the dictionary
                except KeyError:
                    aa_allele = sample.alleles.aa_profile[gene]
                    if aa_allele != '0':
                        allele_notes_dict[int(allele)] = [aa_allele]
                        self.update_aa_allele_notes(
                            gene=gene,
                            allele_notes_dict=allele_notes_dict
                        )

    def read_aa_profile_notes(self):
        """
        Read in all the notes from the profile file (if it exists)
        :return: aa_profile_dict: Dictionary of nucleotide seq type: amino acid sequence type(s)
        """
        aa_profile_dict = dict()
        try:
            with open(self.aa_profile_notes, 'r', encoding='utf-8') as profile:
                # Ignore the header
                _ = profile.readline()
                # Read in all the lines
                for line in profile:
                    # Split the nucleotide sequence type from the amino acid sequence type(s)
                    nt_profile, aa_profiles = line.rstrip().split('\t')
                    # Create a list of all the amino acid sequence types
                    aa_profile_dict[nt_profile] = aa_profiles.split(';')
        except FileNotFoundError:
            pass
        return aa_profile_dict

    def update_aa_profile_notes(self, aa_profile_dict):
        """
        Update the profile file with a novel entry
        :param aa_profile_dict: Dictionary of nucleotide sequence type: amino acid sequence type(s)
        """
        # Overwrite the previous notes file
        with open(self.aa_profile_notes, 'w', encoding='utf-8') as profile:
            # Create the header for the profile file
            profile.write('nt_profile\taa_profile(s)\n')
            # Iterate through all the sequence type entries in the dictionary
            for nt_profile, aa_profiles in aa_profile_dict.items():
                aa_profiles = ';'.join([str(profile) for profile in aa_profiles])
                # Write each nucleotide profile: aa_profile link
                profile.write(f'{nt_profile}\t{aa_profiles}\n')

    def read_aa_allele_notes(self, gene):
        """
        Read in the allele notes file. If it doesn't exist, create it
        :param gene: Name of the current gene being analysed
        :return: Dictionary of nucleotide allele: list of corresponding amino acid alleles
        """
        allele_notes_dict = {}
        # Set the name of the notes file
        notes_file = os.path.join(self.aa_notes_path, f'{gene}_allele_notes.tsv')
        # Attempt to read the notes file
        try:
            with open(notes_file, 'r', encoding='utf-8') as notes:
                # Ignore the header
                _ = notes.readline()
                # Extract the nt allele, ';'-separated list of amino acid alleles from the line
                for line in notes:
                    nt_allele, aa_alleles = line.rstrip().split('\t')
                    # Create a list of all the amino acid alleles
                    allele_notes_dict[int(nt_allele)] = aa_alleles.split(';')
        # Create the file if it doesn't exist
        except FileNotFoundError:
            # Iterate through all the entries in the dictionary of gene name: nt allele: aa allele
            for _, allele_dict in self.aa_nt_allele_link_dict.items():
                for nt_allele, aa_allele in allele_dict.items():
                    # Split the entries on underscores
                    _, nt_allele_number = nt_allele.rsplit('_', 1)
                    _, aa_allele_number = aa_allele.rsplit('_', 1)
                    # Update the notes dictionary with int(nt allele): list(aa allele)
                    allele_notes_dict[int(nt_allele_number)] = [aa_allele_number]
        return allele_notes_dict

    def update_aa_allele_notes(self, gene, allele_notes_dict):
        """
        Update the amino acid allele notes file with the novel alleles
        :param gene: Current gene being analysed
        :param allele_notes_dict: Dictionary of nucleotide allele: list of corresponding amino 
        acid alleles
        """
        # Set the name and path of the notes file
        notes_file = os.path.join(self.aa_notes_path, f'{gene}_allele_notes.tsv')
        with open(notes_file, 'w', encoding='utf-8') as notes:
            # Create the header for the notes file
            notes.write('nt_allele\taa_alleles\n')
            for nt_allele, aa_alleles in sorted(allele_notes_dict.items()):
                # Write the nucleotide allele and list of corresponding amino acid alleles
                aa_alleles = ';'.join(str(allele) for allele in aa_alleles)
                notes.write(f'{str(nt_allele)}\t{aa_alleles}\n')

    def __init__(self, path, amino_acid):
        self.path = pathfinder(path=path)
        self.allele_path = os.path.join(self.path, 'nt_alleles')
        self.aa_allele_path = os.path.join(self.path, 'aa_alleles')
        self.profile_path = os.path.join(self.path, 'nt_profile')
        self.aa_profile_path = os.path.join(self.path, 'aa_profile')
        make_path(self.profile_path)
        self.profile_file = os.path.join(self.profile_path, 'profile.txt')
        self.aa_profile_file = os.path.join(self.aa_profile_path, 'profile.txt')
        self.query_path = os.path.join(self.path, 'query')
        self.report_path = os.path.join(self.path, 'reports')
        make_path(self.report_path)
        novel_alleles = glob(os.path.join(self.report_path, '*.fasta'))
        for novel_allele in novel_alleles:
            os.remove(novel_allele)
        self.aa_notes_path = os.path.join(self.path, 'aa_notes')
        make_path(self.aa_notes_path)
        self.aa_profile_notes = os.path.join(self.aa_notes_path, 'aa_profile_notes.tsv')
        self.amino_acid = amino_acid
        if not self.amino_acid:
            self.combined_targets = os.path.join(self.allele_path, 'combinedtargets.fasta')
        else:
            self.combined_targets = os.path.join(self.aa_allele_path, 'combinedtargets.fasta')
        self.gene_names = []
        self.runmetadata = MetadataObject()
        self.runmetadata.samples = []
        self.cpus = multiprocessing.cpu_count() - 1
        self.profile_report = os.path.join(self.report_path, 'nt_profiles.tsv')
        self.aa_profile_report = os.path.join(self.report_path, 'aa_profiles.tsv')
        try:
            os.remove(self.profile_report)
        except FileNotFoundError:
            pass
        # Fields used for custom outfmt 6 BLAST output:
        self.fieldnames = ['query_id', 'subject_id', 'identical', 'mismatches', 'gaps',
                           'evalue', 'bit_score', 'query_length', 'subject_length', 'alignment_length',
                           'query_start', 'query_end', 'subject_start', 'subject_end',
                           'query_sequence', 'subject_sequence']
        self.extended_fieldnames = self.fieldnames.copy()
        self.extended_fieldnames.insert(14, 'percent_match')
        self.outfmt = '6 qseqid sseqid nident mismatch gaps evalue bitscore qlen slen length ' \
                      'qstart qend sstart send qseq sseq'
        # A string of the header to use for formatting the profile file, and the report headers
        self.data = str()
        self.aa_allele_dict = {}
        self.aa_nt_allele_link_dict = {}


def cli():
    """
    Collect the arguments, create an object, and run the script
    """
    # Parser for arguments
    parser = ArgumentParser(
        description='Determines profiles of strains against previously calculated allele database '
                    'and profile. Creates and/or updates both the database of allele definitions '
                    'and the profile based on novel alleles and/or profiles discovered'
    )
    parser.add_argument(
        '-p', '--path',
        required=True,
        help='Specify path. Note that due to code reuse, the query sequence files must be in the '
             '"query" sub-folder, nucleotide alleles must be in the "nt_alleles" sub-folder, the nucleotide profile '
             'must be named profile.txt and be located in the "nt_profile" sub-folder, amino acid alleles must be in '
             'the "aa_allele" sub-folder, and the amino acid profile must be named profile.txt and be located in the'
             '"aa_profile" sub-folder '
    )
    parser.add_argument(
        '-aa', '--amino_acid',
        action='store_true',
        help='The query sequences are protein.'
    )
    # Get the arguments into an object
    arguments = parser.parse_args()
    SetupLogging(debug=True)
    # Run the profiling pipeline
    updater = Updater(
        path=arguments.path,
        amino_acid=arguments.amino_acid
    )
    updater.main()
    logging.info('Allele Updating complete')


if __name__ == '__main__':
    cli()



#!/usr/bin/env python

"""
Reduce Enterobase wgMLST profiles with a supplied list of genes
"""

# Standard imports
from argparse import ArgumentParser
from csv import DictReader
import logging
import sys
import os

# Third-party imports
from olctools.accessoryFunctions.accessoryFunctions import \
    make_path, \
    SetupLogging
from allele_tools.methods import pathfinder


class ProfileReduce:

    """
    Reduce Enterobase wgMLST profiles
    """

    def main(self):
        """
        Run the required methods in the appropriate order
        """
        self.read_names()
        self.read_profile()

    def read_names(self):
        """
        Read in all the names of the genes of interest
        """
        logging.info('Reading names file')
        with open(self.name_file, 'r', encoding='utf-8') as names:
            self.names = sorted([name.rstrip() for name in names.readlines()])
        logging.debug('Genes used to reduce profile are: %s', '\t'.join(self.names))

    def read_profile(self):
        """
        Load, parse, and reduce the profile information
        """
        logging.info('Reducing profiles')
        with open(self.profile, 'r', encoding='utf-8') as profile:
            with open(self.reduced_profile, 'w', encoding='utf-8') as reduced:
                with open(self.notes_file, 'w', encoding='utf-8') as notes:
                    # Write the header for the notes field
                    notes.write('OriginalSequenceType\tReducedSequenceType\tNotes\n')
                    # Create the header for the reduced profile
                    gene_names = '\t'.join(self.names)
                    reduced_data = f'ST\t{gene_names}\n'
                    logging.info('Loading profile into memory')
                    # Use to DictReader to load the profile file
                    profile_dict = DictReader(profile, dialect='excel-tab')
                    # Initialise a dictionary to store the string of allele numbers: sequence type
                    profile_st = {}
                    logging.info('Parsing profile')
                    # Iterate through all the sequence types in the profile
                    for allele_dict in profile_dict:
                        # Extract the sequence type from the 'ST' column
                        seq_type = allele_dict['ST']
                        # Initialise variables to store the allele numbering information
                        allele_list = []
                        allele_str = ''
                        # Iterate through all the genes of interest
                        for gene in self.names:
                            try:
                                allele_str += allele_dict[gene]
                            # If an allele has been filtered based on length, the sequence type needs to be removed
                            except TypeError:
                                continue
                            # Add the allele number to the list, and to the string
                            allele_list.append(allele_dict[gene])
                        # Check if the string of allele numbers is already in the dictionary
                        if allele_str not in profile_st:
                            # Update the dictionary with the string of alleles: sequence type
                            profile_st[allele_str] = allele_dict[self.names[0]]
                            alleles = '\t'.join(allele_list)
                            # Add the sequence type allele numbers for this sequence to the string
                            reduced_data += f'{seq_type}\t{alleles.rstrip()}\n'
                            # Updated the notes with the sequence type linking information
                            notes.write(f'{seq_type}\t{seq_type}\n')
                        # Reduced profile already exists
                        else:
                            # Extract the original sequence type with this string of allele numbers
                            original_seq_type = profile_st[allele_str]
                            # Write the sequence type linking information, making note of the fact
                            # that this is a duplicate
                            notes.write(f'{seq_type}\t{original_seq_type}\tduplicate\n')
                    # Write the reduced profile information to file
                    reduced.write(reduced_data)

    def __init__(self, profile, names, output='profile'):
        logging.info('Welcome to profile reducer!')
        self.profile = pathfinder(path=profile)
        try:
            assert os.path.isfile(self.profile)
        except AssertionError as exc:
            logging.error('Cannot locate the specified profile file: %s', self.profile)
            raise SystemExit from exc
        # Create the folder into which the reduced profile and notes are to be placed
        self.report_path = os.path.join(os.path.dirname(self.profile), output)
        make_path(self.report_path)
        self.reduced_profile = os.path.join(self.report_path, 'profile.txt')
        self.notes_file = os.path.join(self.report_path, 'reducing_notes.txt')
        self.name_file = pathfinder(path=names)
        try:
            assert os.path.isfile(self.name_file)
        except AssertionError as exc:
            logging.error('Cannot find the supplied file with gene names: %s', self.name_file)
            raise SystemExit from exc
        self.names = []
        self.profile_dict = {}
        self.allele_dict = {}


def cli():
    """
    Collect the arguments, create an object, and run the script
    """
    # Parser for arguments
    parser = ArgumentParser(description='Extract the genes of interest from a profile file')
    parser.add_argument(
        '-p', '--profile',
        metavar='profile',
        required=True,
        help='Name and path of profile file.')
    parser.add_argument(
        '-n', '--names',
        metavar='names',
        required=True,
        help='Name and path to a file containing the gene names (one per line) to be extracted '
             'from the profile')
    # Get the arguments into an object
    arguments = parser.parse_args()
    SetupLogging(debug=True)
    reduce = ProfileReduce(
        profile=arguments.profile,
        names=arguments.names
    )
    reduce.main()
    logging.info('Profile reduction complete!')
    # Prevent the arguments being printed to the console (they are returned in order for the tests to work)
    sys.stderr = open(os.devnull, 'w', encoding='utf-8')
    return arguments


if __name__ == '__main__':
    cli()



#!/usr/bin/env python

"""
Unit and integration tests for allele_tools/stec.py allele_split
"""

# Standard imports
from unittest.mock import patch
from glob import glob
import argparse
import shutil
import os

# Third-party imports
from Bio import SeqIO
import pytest

# Local imports
from allele_tools.stec import (
    allele_split,
    cli
)


@pytest.fixture(name='variables', scope='module')
def setup():
    class Variables:
        def __init__(self):
            # Extract the connection string prior to running tests
            self.test_path = os.path.abspath(os.path.dirname(__file__))
            self.file_path = os.path.join(self.test_path, 'test_files', 'stec_allele_split')
            self.original_file_path = os.path.join(self.file_path, 'original_files')
            self.original_allele_files = glob(os.path.join(self.original_file_path, '*.fasta'))
            self.query_path = os.path.join(self.file_path, 'query')
            self.output_path = os.path.join(self.file_path, 'split_alleles')

    return Variables()


def prepare_files(variables):
    if not os.path.isdir(variables.query_path):
        os.makedirs(variables.query_path)
    for allele_file in variables.original_allele_files:
        shutil.copyfile(
            src=allele_file,
            dst=os.path.join(variables.query_path, os.path.basename(allele_file))
        )


def clean_outputs(variables):
    if os.path.isdir(variables.query_path):
        shutil.rmtree(variables.query_path)
    if os.path.isdir(variables.output_path):
        shutil.rmtree(variables.output_path)


@patch('argparse.ArgumentParser.parse_args')
def test_profile_reduce_missing_tilde_profile_missing_files(mock_args, variables):
    with pytest.raises(SystemExit):
        mock_args.return_value = argparse.Namespace(
            query_path=variables.query_path,
            output_path=variables.output_path
        )
        arguments = cli()
        allele_split(args=arguments)


@patch('argparse.ArgumentParser.parse_args')
def test_profile_reduce_missing_tilde_profile(mock_args, variables):
    prepare_files(variables=variables)
    mock_args.return_value = argparse.Namespace(
        query_path=variables.query_path,
        output_path=variables.output_path
    )
    arguments = cli()
    allele_split(args=arguments)
    variables.output_alleles = sorted(glob(os.path.join(variables.output_path, '*.fasta')))
    assert len(variables.output_alleles) == 1096


def test_allele_contents(variables):
    for record in SeqIO.parse(variables.output_alleles[0], 'fasta'):
        assert record.id == 'ECs1205_1'
        assert str(record.seq) == \
               'MKCILFKWVLCLLLGFSSVSYSREFTIDFSTQQSYVSSLNSIRTEISTPLEHISQGTTSV' \
               'SVINHTPPGSYFAVDIRGLDVYQARFDHLRLIIEQNNLYVAGFVNTATNTFYRFSDFTHI' \
               'SVPGVTTVSMTTDSSYTTLQRVAALERSGMQISRHSLVSSYLALMEFSGNTMTRDASRAV' \
               'LRFVTVTAEALRFRQIQREFRQALSETAPVYTMTPGDVDLTLNWGRISNVLPEYRGEDGV' \
               'RVGRISFNNISAILGTVAVILNCHHQGARSVRAVNEESQPECQITGDRPVIKINNTLWES' \
               'NTAAAFLNRKSQFLYTTGK*'


def test_clean_outputs(variables):
    clean_outputs(variables=variables)
    assert not os.path.isdir(variables.query_path)
    assert not os.path.isdir(variables.output_path)



#!/usr/bin/env python

"""
Unit and integration tests for allele_tools/allele_translate_reduce.py
"""

# Standard imports
from unittest.mock import patch
from glob import glob
import argparse
import shutil
import os

# Third-party imports
from Bio import SeqIO
import pytest

# Local imports
from allele_tools.allele_updater import \
    Updater, \
    cli


@pytest.fixture(name='variables', scope='module')
def setup():
    class Variables:
        def __init__(self):
            # Extract the connection string prior to running tests
            self.test_path = os.path.abspath(os.path.dirname(__file__))
            self.file_path = os.path.join(self.test_path, 'test_files', 'allele_updater')
            self.original_files_path = os.path.join(self.file_path, 'original_files')
            self.original_nt_allele_files = glob(os.path.join(self.original_files_path, 'nt_alleles', 'ECs12*'))
            self.original_nt_profile_file = os.path.join(self.original_files_path, 'nt_profile', 'profile.txt')
            self.original_nt_query_files = glob(os.path.join(self.original_files_path, 'nt_query', '*.fasta'))
            self.original_aa_allele_files = glob(os.path.join(self.original_files_path, 'aa_alleles', 'ECs12*'))
            self.original_aa_profile_path = os.path.join(self.original_files_path, 'aa_profile')
            self.original_aa_profile_file = os.path.join(self.original_aa_profile_path, 'profile.txt')
            self.original_aa_query_files = glob(os.path.join(self.original_files_path, 'aa_query', '*.fasta'))
            self.nt_allele_path = os.path.join(self.file_path, 'nt_alleles')
            self.nt_profile_file = os.path.join(self.file_path, 'nt_profile', 'profile.txt')
            self.nt_profile_path = os.path.join(self.file_path, 'nt_profile')
            self.aa_allele_path = os.path.join(self.file_path, 'aa_alleles')
            self.aa_profile_path = os.path.join(self.file_path, 'aa_profile')
            self.aa_profile_file = os.path.join(self.aa_profile_path, 'profile.txt')
            self.aa_notes_path = os.path.join(self.file_path, 'aa_notes')
            self.query_path = os.path.join(self.file_path, 'query')
            self.report_path = os.path.join(self.file_path, 'reports')
            self.aa_report_file = os.path.join(self.report_path, 'aa_profiles.tsv')

            self.report_file = os.path.join(self.report_path, 'nt_profiles.tsv')
            self.length_dict = {
                'ECs2973': 90,
                'ECs2974': 316,
                'ECs1205': 316,
                'ECs1206': 88
            }
            self.fake_path = os.path.join('~', 'completely_fake_path')

    return Variables()


def clean_outputs(variables):
    if os.path.isdir(variables.nt_allele_path):
        shutil.rmtree(variables.nt_allele_path)
    if os.path.isdir(variables.nt_profile_path):
        shutil.rmtree(variables.nt_profile_path)
    if os.path.isdir(variables.aa_allele_path):
        shutil.rmtree(variables.aa_allele_path)
    if os.path.isdir(variables.aa_profile_path):
        shutil.rmtree(variables.aa_profile_path)
    if os.path.isdir(variables.query_path):
        shutil.rmtree(variables.query_path)
    if os.path.isdir(variables.report_path):
        shutil.rmtree(variables.report_path)
    if os.path.isdir(variables.aa_notes_path):
        shutil.rmtree(variables.aa_notes_path)


def prepare_files(variables, nucleotide=True, links=False):
    if not os.path.isdir(variables.nt_allele_path):
        os.makedirs(name=variables.nt_allele_path)
    for nt_allele in variables.original_nt_allele_files:
        shutil.copyfile(
            src=nt_allele,
            dst=os.path.join(variables.nt_allele_path, os.path.basename(nt_allele))
        )
    if not os.path.isdir(variables.nt_profile_path):
        os.makedirs(name=variables.nt_profile_path)
    shutil.copyfile(
        src=variables.original_nt_profile_file,
        dst=os.path.join(variables.nt_profile_path, os.path.basename(variables.original_nt_profile_file))
    )
    if not os.path.isdir(variables.aa_allele_path):
        os.makedirs(name=variables.aa_allele_path)
    for aa_allele in variables.original_aa_allele_files:
        shutil.copyfile(
            src=aa_allele,
            dst=os.path.join(variables.aa_allele_path, os.path.basename(aa_allele))
        )
    if not os.path.isdir(variables.aa_profile_path):
        os.makedirs(name=variables.aa_profile_path)
    shutil.copyfile(
        src=variables.original_aa_profile_file,
        dst=os.path.join(variables.aa_profile_path, os.path.basename(variables.original_aa_profile_file))
    )
    if links:
        link_file = 'aa_nt_profile_links.tsv'
        shutil.copyfile(
            src=os.path.join(variables.original_aa_profile_path, link_file),
            dst=os.path.join(variables.aa_profile_path, link_file)
        )
    os.makedirs(name=variables.query_path)
    if nucleotide:
        for nt_query in variables.original_nt_query_files:
            shutil.copyfile(
                src=nt_query,
                dst=os.path.join(variables.query_path, os.path.basename(nt_query))
            )
    else:
        for aa_query in variables.original_aa_query_files:
            shutil.copyfile(
                src=aa_query,
                dst=os.path.join(variables.query_path, os.path.basename(aa_query))
            )


@patch('argparse.ArgumentParser.parse_args')
def test_allele_updater_integration(mock_args, variables):
    prepare_files(variables=variables)
    mock_args.return_value = argparse.Namespace(
        path=variables.file_path,
        amino_acid=False,
    )
    cli()
    assert os.path.isfile(variables.report_file)


def test_report_contents(variables):
    variables.report_contents = open(variables.report_file, 'r', encoding='utf-8').readlines()
    assert variables.report_contents[1] == '2013-SEQ-0039	259156	6	8\n'


def test_novel_alleles(variables):
    novel_allele_file = os.path.join(variables.report_path, 'nt_ECs1205_novel_alleles.fasta')
    novel_alleles = open(novel_allele_file, 'r', encoding='utf-8').readlines()
    assert novel_alleles[0] == '>ECs1205_1000000\n'
    clean_outputs(variables=variables)


def test_tilde_path(variables):
    Updater(
        path=variables.fake_path,
        amino_acid=False
    )
    abs_path = os.path.abspath(os.path.expanduser(os.path.join(variables.fake_path)))
    assert os.path.isdir(abs_path)
    shutil.rmtree(abs_path)


@patch('argparse.ArgumentParser.parse_args')
def test_allele_updater_integration_aa_query(mock_args, variables):
    prepare_files(
        variables=variables,
        nucleotide=False
    )
    mock_args.return_value = argparse.Namespace(
        path=variables.file_path,
        amino_acid=True,
    )
    cli()
    assert os.path.isfile(variables.aa_report_file)


def test_aa_report_contents(variables):
    aa_report_contents = open(variables.aa_report_file, 'r', encoding='utf-8').readlines()
    assert aa_report_contents[2] == 'ECs1205_11\t175\t11\t0\n'
    clean_outputs(variables=variables)


@patch('argparse.ArgumentParser.parse_args')
def test_allele_updater_integration_new_aa_files(mock_args, variables):
    prepare_files(variables=variables)
    shutil.rmtree(variables.aa_allele_path)
    mock_args.return_value = argparse.Namespace(
        path=variables.file_path,
        amino_acid=False,
    )
    cli()
    assert os.path.isfile(variables.report_file)


def test_novel_aa_alleles(variables):
    novel_aa_allele_file = os.path.join(variables.aa_allele_path, 'ECs1206_alleles.fasta')
    records = list(SeqIO.parse(novel_aa_allele_file, 'fasta'))
    assert records[0].seq == \
           'MKKMFMAVLFALASVNAMAADCAKGKIEFSKYNEDDTFTVKVDGKEYWTSRWNLQPLLQSAQLTGMTVTIKSSTCESGSGFAEVQFNND*'
    clean_outputs(variables=variables)



#!/usr/bin/env python

"""
Translate nucleotide alleles to amino acid, and remove duplicates
"""

# Standard imports
from argparse import ArgumentParser
from glob import glob
import logging
import shutil
import sys
import os

# Third-party imports
from olctools.accessoryFunctions.accessoryFunctions import make_path, SetupLogging
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO

# Local imports
from allele_tools.allele_profiler import read_profile
from allele_tools.profile_reduce import ProfileReduce
from allele_tools.methods import \
    evaluate_translated_length, \
    generic_evaluate_translated_length, \
    remove_combined_db_files, \
    pathfinder


class Translate:

    """
    Translate and reduce alleles
    """

    def main(self):
        """
        Run the appropriate methods in the correct order
        """
        # Read in all the allele files
        self.load_alleles()
        # Parse the alleles
        self.parse_alleles(length_dict=self.length_dict)
        # If a profile file has been provided, run the profiling methods
        if self.profile_file:
            # Extract seq_type:allele comprehensions from the nucleotide profile file
            self.profile_data = read_profile(profile_file=self.profile_file)
            # Create the amino acid profiles
            self.aa_profile()
            # Remove duplicate profiles
            reduce = ProfileReduce(
                profile=self.aa_profile_file,
                names=self.gene_name_file
            )
            reduce.main()
            # Read in the profile data again now that the profile file has been updated
            self.aa_profile_data = read_profile(profile_file=self.aa_profile_file)
            # Find linkages between nucleotide and amino acid profiles
            self.profile_link()
            # Write profile links to file
            self.link_file()
            # Copy and rename the reduced profile file
            self.copy_profile()

    def load_alleles(self):
        """
        Use SeqIO to read in all the gene sequences
        """
        for allele_file in self.sequence_files:
            gene_name = os.path.splitext(os.path.basename(allele_file))[0]
            # Add the gene name to the set of names
            self.gene_names.add(gene_name)
            self.allele_dict[gene_name] = SeqIO.to_dict(SeqIO.parse(allele_file, 'fasta'))

    def parse_alleles(self, length_dict=None):
        """
        Parse the allele files to translate the amino acid sequence using BioPython. Write the
        amino acid sequence to file. Store the allele name in the notes. Find duplicates, and link
        the nucleotide allele name to the amino acid allele name
        :param length_dict: Dictionary of gene name: minimum acceptable length of translated sequence
        """
        logging.info('Translating and parsing alleles')
        for gene_name, allele_dict in self.allele_dict.items():
            # Initialise the dictionary to store the links between the nt alleles and the aa
            # alleles with the name of the gene
            self.allele_links[gene_name] = {}
            #  Initialise a set of allele that do not pass the required filters, and
            # must be removed from the database
            remove_list = set()
            logging.info('Processing %s', gene_name)
            # Initialise a dictionary to store the translated allele sequence: allele name
            seq_allele = {}
            # Set the name and path of the allele and notes files
            allele_file = os.path.join(self.translated_path, f'{gene_name}.fasta')
            notes_file = os.path.join(self.notes_path, f'{gene_name}_notes.txt')
            # Open the file to store the translated alleles
            with open(allele_file, 'w', encoding='utf-8') as aa_alleles:
                # Open the notes file
                with open(notes_file, 'w', encoding='utf-8') as notes:
                    # Create the header for the notes file
                    notes.write('nt_allele\taa_allele\tnote\n')
                    # Iterate through all the alleles in the dictionary
                    for allele, details in allele_dict.items():
                        # Create a list to store notes
                        note = []
                        # Create a boolean to track whether the sequence is filtered
                        filtered = False
                        # Create a string to store the untrimmed nucleotide sequence
                        original_nt_sequence = str(details.seq)
                        # Calculate the translated sequence
                        translated_allele = details.seq.translate()
                        # Remove all sequence after a stop codon (*)
                        split_allele = translated_allele.split('*')
                        # Create a string to hold the trimmed sequence
                        trimmed_seq = str()
                        # If there are multiple stop codons in the sequence, trim to the first one
                        if len(split_allele) > 2:
                            if split_allele[1]:
                                for trimmed in split_allele[1:]:
                                    trimmed_seq += f'*{trimmed}'
                            else:
                                for trimmed in split_allele[2:]:
                                    trimmed_seq += f'*{trimmed}'
                            note.append(f'Trimmed {trimmed_seq} from end')
                        elif len(split_allele) == 1:
                            pass
                        else:
                            if split_allele[-1] and not str(translated_allele).endswith('*'):
                                trimmed_seq += split_allele[-1]
                                note.append(f'Trimmed {trimmed_seq} from end')
                        # Create a dictionary to store the allele nt sequences to check to see if
                        # any are duplicates following trimming
                        nt_sequences = {}
                        filtered, note, nt_sequences, translated_allele = Translate.trim_alleles(
                            note=note,
                            allele=allele,
                            sequence=details,
                            gene_name=gene_name,
                            nt_allele_path=self.path,
                            trim_length=len(trimmed_seq),
                            length_dict=length_dict,
                            filtered=filtered,
                            nt_sequences=nt_sequences,
                            original_nt_sequence=original_nt_sequence
                        )
                        # If the allele has not been filtered through the trim_alleles
                        if not filtered:
                            # Determine if this amino acid allele is new
                            if str(translated_allele) not in seq_allele:
                                # Add the string of the amino acid sequence to the dictionary
                                seq_allele[str(translated_allele)] = allele
                                # Create a SeqRecord of the translated allele
                                seq_record = SeqRecord(
                                    seq=translated_allele,
                                    id=allele,
                                    name=str(),
                                    description=str()
                                )
                                # Write the SeqRecord to file
                                SeqIO.write(
                                    sequences=seq_record,
                                    handle=aa_alleles,
                                    format='fasta'
                                )
                                # Update the notes with the allele naming information
                                notes.write(f'{allele}\t{allele}\t{";".join(note)}\n')
                                # Populate the linking dictionary with the nt allele: aa allele
                                self.allele_links[gene_name][allele.split('_')[-1]] = \
                                    allele.split('_')[-1]
                            # Amino acid allele already exists
                            else:
                                # Extract the allele name corresponding to the translated sequence
                                aa_allele = seq_allele[str(translated_allele)]
                                # Update the notes, including that this allele is a duplicate, and a
                                # pointer to the original
                                notes.write(
                                    f'{allele}\t{aa_allele}\tDuplicate'
                                )
                                if not note:
                                    notes.write('\n')
                                else:
                                    notes.write(f'; {";".join(note)}\n')
                                # Populate the linking dictionary with the nt allele: aa allele
                                self.allele_links[gene_name][allele.split('_')[-1]] = \
                                    aa_allele.split('_')[-1]
                                self.allele_links[gene_name]['0'] = '0'
                        # Filtered alleles must be removed from the database
                        else:
                            Translate.write_filtered_allele_notes(
                                notes=notes,
                                allele=allele,
                                note=note,
                            )
                            remove_list.add(allele)
                            # Create a SeqRecord of the translated allele
                            seq_record = SeqRecord(
                                seq=translated_allele,
                                id=allele,
                                name=str(),
                                description=str()
                            )
                            # Write the filtered alleles to the filtered alleles file
                            Translate.create_or_update_filtered_files(
                                gene=gene_name,
                                allele_path=self.translated_path,
                                records_filter=[seq_record]
                            )
            # Remove the filtered sequences from the database
            Translate.remove_filtered_alleles_from_db(
                gene_name=gene_name,
                allele_list=remove_list,
                nt_allele_path=self.path,
            )
            self.load_alleles()

    @staticmethod
    def trim_alleles(note, allele, sequence, gene_name, nt_allele_path, trim_length, length_dict, filtered,
                     nt_sequences, original_nt_sequence):
        """
        Trim the alleles based on location of stop codons and whether the sequence is a multiple of three nucleotides.
        Evaluate the trimmed sequence based on length and contents.
        :param note: List of sequence-specific notes
        :param allele: String of the allele identifier
        :param sequence: SeqIO sequence object of nucleotide sequence
        :param gene_name: String of the gene name to which the allele corresponds
        :param nt_allele_path: String of the absolute path to the folder in which the nucleotide alleles are located
        :param trim_length: Integer of the number of nucleotides to be trimmed (due to internal stop codons)
        :param length_dict: Dictionary of minimum acceptable length for each gene in the analysis
        :param filtered: Boolean to track whether the sequence fails the quality/length checks
        :param nt_sequences: Dictionary of allele: sequence
        :param original_nt_sequence: String of the untrimmed nucleotide sequence of the allele
        :return: filtered: Updated boolean of whether the sequence fails quality/length checks
        :return: note: Update list of sequence-specific notes
        :return: nt_sequences: Updated dictionary of allele: sequence
        :return: translated_allele: SeqIO sequence object of trimmed, translated allele
        """
        # Determine the length of sequence to trim from the end of the sequence
        # Multiply the number of amino acid residues to trim by three to get the number of nucleotides
        # Add the modulo three of the sequence to yield a final length divisible by three
        nt_to_trim = 3 * trim_length + len(sequence) % 3
        # Check if trimming is required
        if nt_to_trim:
            # Slice the string by the number of required bases at the end
            sequence = sequence[:-nt_to_trim]
            # Update the nucleotide sequence in the database with the trimmed version
            Translate.update_allele_db(
                nt_allele_path=nt_allele_path,
                gene=gene_name,
                allele=allele,
                nt_sequence=sequence
            )
        # Translate the sequence to amino acid
        translated_allele = sequence.seq.translate()
        # Perform content and length checks of the protein sequence
        if length_dict:
            filtered, note = evaluate_translated_length(
                aa_seq=str(translated_allele),
                length_dict=length_dict,
                gene=gene_name,
                notes=note,
                filtered=filtered
            )
        else:
            filtered, note = generic_evaluate_translated_length(
                aa_seq=str(translated_allele),
                sequence=original_nt_sequence,
                gene=gene_name,
                notes=note,
                filtered=filtered,
                cutoff=0.95
            )
        # Search the amino acid database for matches
        filtered, note, nt_sequences = Translate.find_duplicates(
            nt_sequences=nt_sequences,
            nt_sequence=sequence,
            allele=allele,
            filtered=filtered,
            note=note
        )
        # Update the notes if the sequence does not end with a stop codon
        if not str(translated_allele).endswith('*'):
            note.append('Trimmed sequence did not end with a stop codon')
        return filtered, note, nt_sequences, translated_allele
    
    @staticmethod
    def update_allele_db(nt_allele_path, gene, allele, nt_sequence):
        """
        Update nucleotide allele files with newly trimmed sequences
        :param nt_allele_path: String of the absolute path to the folder containing the nucleotide allele database
        :param gene: String of the name of the current gene
        :param allele: String of the allele header (geneName_alleleID)
        :param nt_sequence: SeqIO object of the nucleotide sequence
        """
        # Set the name of the allele database file to update by joining the allele path and the gene name
        nt_allele_file = os.path.join(nt_allele_path, f'{gene}.fasta')
        # Create a list to store all database records (including modified ones)
        updated_records = []
        # Iterate over all the sequences in the nucleotide allele file
        for record in SeqIO.parse(handle=nt_allele_file, format='fasta'):
            # Check if the header of the database sequence matches the header of the query sequence
            if record.id == allele:
                # Update the record to be the query allele sequence object
                record = nt_sequence
            # Append the record to the list of records
            updated_records.append(record)
        # Overwrite the allele file with the records
        SeqIO.write(updated_records, handle=nt_allele_file, format='fasta')
    
    @staticmethod
    def find_duplicates(nt_sequences, nt_sequence, allele, filtered, note):
        """
        Match a query amino acid sequence against the protein allele database file
        :param nt_sequences: Dictionary of allele: sequence
        :param nt_sequence: SeqIO sequence object of the trimmed nucleotide sequence
        :param allele: String of the allele header (geneName_alleleID)
        :param filtered: Boolean of whether the sequence passes content/length filters
        :param note: List of sequence-specific notes
        :return: filtered: Updated boolean of sequence-specific content/length filtering results
        :return: note: Updated list of sequence-specific notes
        :return: nt_sequences: Updated list of allele: sequence
        """
        # Initialise a list to store any matches
        matches = []
        # Iterate over allele header, sequence in the dictionary
        for prev_allele, nt_seq in nt_sequences.items():
            # Check if the current nucleotide sequence matches a previous entry
            if nt_sequence == nt_seq:
                # Append matches to the list
                matches.append(prev_allele)
        # Check if the list has been populated with matches
        if not matches:
            # If no matches, update the dictionary with the novel sequence
            nt_sequences[allele] = nt_sequences
            return filtered, note, nt_sequences
        # If there are matches
        for match in matches:
            # Set the filtering boolean to True (this is not a novel sequence, so do not update the database)
            filtered = True
            # Update the note
            note.append(f'Trimmed nt sequence matches previous allele sequence: {match}')
        return filtered, note, nt_sequences

    @staticmethod
    def write_filtered_allele_notes(notes, allele, note):
        """
        Write notes for filtered sequences to file
        :param notes: Filehandle for notes file
        :param allele: String of the allele header (geneName_alleleID)
        :param note: List of sequence-specific notes
        """
        # Write the allele \t ND (because the sequence is filtered,there is no amino acid allele) \t
        notes.write(f'{allele}\tND\tFiltered: {"; ".join(note)}\n')

    @staticmethod
    def remove_filtered_alleles_from_db(gene_name, allele_list, nt_allele_path):
        """
        Remove filtered sequences from the allele database file
        :param gene_name: String of the gene name currently being analysed
        :param allele_list: List of alleles to be removed from the database
        :param nt_allele_path: String of the absolute path to the folder containing the nucleotide allele database
        """
        # Remove the combinedtargets.fasta and BLAST database files from the folder
        remove_combined_db_files(
            allele_path=nt_allele_path
        )
        # Set the name and path of the allele file
        nt_allele_file = os.path.join(nt_allele_path, f'{gene_name}.fasta')
        # Create a list of all the records in the database file using SeqIO
        records = SeqIO.parse(nt_allele_file, 'fasta')
        # Initialise a list to store unfiltered records
        records_keep = []
        # Initialise a list to store filtered records
        records_filter = []
        # Iterate over all the records in the database
        for record in records:
            # Check if the allele header matches any of the headers of alleles to be filtered
            if record.id not in allele_list:
                # If the record is not to be filtered, add it to the keep list
                records_keep.append(record)
            # If the record header matches any of the alleles to be filtered, add it to the filtered list
            else:
                records_filter.append(record)
        # Overwrite the nucleotide allele database file with all the unfiltered records
        with open(nt_allele_file, 'w', encoding='utf-8') as allele_file:
            for record in records_keep:
                SeqIO.write(
                    sequences=record,
                    handle=allele_file,
                    format='fasta'
                )
        # Write the filtered records to the filtered records file
        if records_filter:
            Translate.create_or_update_filtered_files(
                gene=gene_name,
                allele_path=nt_allele_path,
                records_filter=records_filter
            )

    @staticmethod
    def create_or_update_filtered_files(gene, allele_path, records_filter):
        """
        Write the filtered alleles to the filtered alleles file
        :param gene: String of the gene name currently being analysed
        :param allele_path: String of the absolute path to the folder containing the allele database
        :param records_filter: List of SeqIO sequence objects for alleles to be added to the filtered alleles file
        """
        # Set the name and path of the filtered alleles file
        filtered_allele_file = os.path.join(allele_path, f'{gene}_filtered.txt')
        # Append the filtered alleles to the file
        with open(filtered_allele_file, 'a+', encoding='utf-8') as filtered_alleles:
            for record in records_filter:
                SeqIO.write(
                    sequences=record,
                    handle=filtered_alleles,
                    format='fasta'
                )

    def aa_profile(self):
        """
        Create the amino acid profile
        """
        # Create a list to store profiles containing alleles that have been quality filtered
        filtered_list = set()
        # Initialise a dictionary to store sequenceType:geneName_alleleID
        filtered_dict = {}
        # Initialise a string to hold the profile
        profile_str = ''
        # Iterate through all the sequence types in the profile_data dictionary
        for seq_type in sorted(int(st) for st in self.profile_data):
            # Create a string to store the allele information
            allele_string = str()
            # Iterate through all the gene: allele entries in the nested dictionary
            for gene_name, allele in self.profile_data[str(seq_type)].items():
                try:
                    # Extract the linked amino acid allele from the dictionary
                    allele_string += self.allele_links[gene_name][str(allele)] + '\t'
                # If the gene name + allele ID is not in the allele link dictionary, add the sequence type to the set
                except KeyError:
                    filtered_list.add(seq_type)
                    # Initialise the sequence type key in the filtered dictionary as required
                    if seq_type not in filtered_dict:
                        filtered_dict[seq_type] = []
                    # Update the dictionary with the filtered geneName_alleleIdentifier
                    filtered_dict[seq_type].append(f'{gene_name}_{allele}')
            if allele_string:
                # Add the sequence type to the profile string
                profile_str += f'{str(seq_type)}\t{allele_string}'
                # Remove trailing whitespace and add a newline for proper formatting
                profile_str = profile_str.rstrip()
                profile_str += '\n'
        # Create the amino acid profile file
        with open(self.aa_profile_file, 'w', encoding='utf-8') as aa_profile:
            # Create a string of tab-delimited gene names to be used in the header
            names = '\t'.join(sorted(list(self.gene_names)))
            # Write the header string
            aa_profile.write(f'ST\t{names.rstrip()}\n')
            # Write the tab-delimited profile string
            aa_profile.write(profile_str)
        # Create the names file
        with open(self.gene_name_file, 'w', encoding='utf-8') as gene_file:
            # Write the names to file
            gene_file.write('\n'.join(sorted(list(self.gene_names))))
        if filtered_list:
            filtered_list = sorted(list(filtered_list))
            # Remove the filtered profiles from the profile files
            Translate.filter_profiles(
                filtered_list=filtered_list,
                profile_file=self.profile_file,
            )
            # Write the notes to file
            Translate.filter_notes(
                filtered_list=filtered_list,
                filtered_dict=filtered_dict,
                profile_path=os.path.dirname(self.profile_file)
            )
            self.profile_data = read_profile(profile_file=self.profile_file)

    @staticmethod
    def filter_profiles(filtered_list, profile_file):
        """
        Remove filtered profiles from the profile file. Write them to the filtered profile file
        :param filtered_list: Set of filtered sequence types
        :param profile_file: String of the absolute path to the profile file
        """
        # Initialise a list to store the filtered sequence types
        filtered_rows = []
        # Initialise a list to store the unfiltered sequence types
        keep_rows = []
        # Read in the profile files
        with open(profile_file, 'r', encoding='utf-8') as profiles:
            # Iterate over all the profiles in the profile files
            for row in profiles:
                # Extract the sequence type from the row
                seq_type = row.split('\t')[0]
                # Check if it is the header, which starts with 'ST'
                if seq_type == 'ST':
                    # Filter the header
                    filtered_rows.append(row)
                # Check to see if the sequence type of the row is present in the list of filtered sequence types
                if str(seq_type) in [str(filtered) for filtered in filtered_list]:
                    # Add the row to the list of filtered rows
                    filtered_rows.append(row)
                # Unfiltered rows are added to the list of unfiltered rows
                else:
                    keep_rows.append(row)
        # Extract the absolute path of the folder in which the profile file is located
        profile_path = os.path.dirname(profile_file)
        # Overwrite the profile file with the unfiltered profiles
        with open(os.path.join(profile_path, 'profile.txt'), 'w', encoding='utf-8') as updated_profile:
            updated_profile.write(''.join(keep_rows))
        # Overwrite the filtered profile file with the filtered profiles
        with open(os.path.join(profile_path, 'filtered_profiles.txt'), 'w', encoding='utf-8') as filtered_profiles:
            filtered_profiles.write(''.join(filtered_rows))

    @staticmethod
    def filter_notes(filtered_list, filtered_dict, profile_path):
        """
        Write the notes regarding profile filtering to file
        :param filtered_list: Set of filtered sequence types
        :param filtered_dict: Dictionary of sequenceType:geneName_alleleID
        :param profile_path: String of the absolute path to the folder containing the profile file
        :return:
        """
        # Set the absolute path of the file containing the profile filtering notes
        filtered_notes = os.path.join(profile_path, 'filtering_notes.txt')
        # Write the notes to file
        with open(filtered_notes, 'w', encoding='utf-8') as notes:
            # Create the header
            notes.write('SequenceType\tFilteringNote\n')
            # Create a new line for each filtered profile
            for seq_type in filtered_list:
                # Create a sting of the list of filtered alleles present in this profile
                gene_allele = ';'.join(filtered_dict[seq_type])
                # Write the sequence type and all the missing alleles to the note
                notes.write(f'{seq_type}\t{gene_allele} missing\n')
    
    def profile_link(self):
        """
        Link nucleotide and amino profiles
        """
        # Initialise a dictionary to store to number of matches between a query profile and the profiles in the database
        match_score = {}
        # Iterate over all the profiles in the profile file
        for seq_type, gene_dict in self.profile_data.items():
            # Initialise the seq_type key in the dictionary
            self.profile_matches[seq_type] = set()
            match_score[seq_type] = {}
            # Iterate over all the gene name, allele ID combinations in the profile dictionary
            for gene, allele in gene_dict.items():
                # Iterate over all the profiles in the amino acid profile file
                for aa_st, aa_gene_dict in self.aa_profile_data.items():
                    # Use the gene name to extract the amino acid allele ID from the profile file.
                    # Also extract the amino acid allele ID from the linking dictionary with the gene name and
                    # nucleotide allele ID. Check if they match
                    if aa_gene_dict[gene] == self.allele_links[gene][allele]:
                        # Initialise the amino acid sequence type in the dictionary
                        if aa_st not in match_score[seq_type]:
                            match_score[seq_type][aa_st] = 0
                        # Increment the number of matches to the profile
                        match_score[seq_type][aa_st] += 1
        # Iterate over all the matches to the profiles in the profile file
        for seq_type, aa_st_dict in match_score.items():
            # Iterate over the amino acid sequence type matches
            for aa_st, matches, in aa_st_dict.items():
                # Check if the number of matches observed is equal to the required number of matches (one for each gene)
                if matches == len(self.gene_names):
                    # Update the dictionary of matches with the amino acid sequence type: nucleotide sequence type
                    self.profile_matches[aa_st].add(seq_type)

    def link_file(self):
        """
        Write linking details between nucleotide and amino acid profiles to file
        """
        #
        with open(self.aa_nt_profile_link_file, 'w', encoding='utf-8') as link:
            # Write the header information
            link.write('aa_seq_type\tnt_seq_types\n')
            # Iterate over all the profile matches in the dictionary
            for seq_type, match_set in self.profile_matches.items():
                # Check the length of the match
                if len(match_set) == 1:
                    # With a single match, convert the set to a string
                    nt_st = ''.join(match_set)
                    link.write(f'{seq_type}\t{nt_st}\n')
                # Multiple profile matches
                else:
                    # Semicolon-join the set of matches. Since they are strings, first typecase to int for proper
                    # sorting before converting back to strings
                    nt_st = ';'.join(str(nt_st) for nt_st in sorted(int(nt_st) for nt_st in match_set))
                    link.write(f'{seq_type}\t{nt_st}\n')

    def copy_profile(self):
        """
        Copy the reduced profile file from the profile folder to the base aa_profile folder
        """
        # Use shutil to copy and rename the profile file to the root of the report path
        shutil.copyfile(
            src=os.path.join(self.report_path, 'profile', 'profile.txt'),
            dst=os.path.join(self.report_path, 'profile.txt')
        )

    def __init__(self, path, profile, report_path=os.path.join(os.getcwd(), 'aa_profile'),
                 translated_path=os.path.join(os.getcwd(), 'aa_alleles'), length_dict=None):

        logging.info('Welcome to the allele translator!')
        self.path = pathfinder(path=path)
        if profile:
            if not os.path.isfile(profile):
                self.profile_file = os.path.join(self.path, 'nt_profile', 'profile.txt')
            else:
                self.profile_file = profile
            try:
                assert os.path.isfile(self.profile_file)
            except AssertionError as exc:
                logging.error(
                    'Cannot locate the required profile file: %s. Please ensure that '
                    'the file name and path of your file is correct', self.profile_file
                )
                raise SystemExit from exc
        else:
            self.profile_file = None
        self.sequence_files = glob(os.path.join(self.path, '*.fasta'))
        try:
            assert self.sequence_files
        except AssertionError as exc:
            logging.error('Could not locate alleles in provided allele path: %s', self.path)
            raise SystemExit from exc
        self.report_path = pathfinder(path=report_path)
        make_path(inpath=self.report_path)
        self.translated_path = pathfinder(path=translated_path)
        make_path(inpath=self.translated_path)
        self.notes_path = os.path.join(self.translated_path, 'notes')
        make_path(inpath=self.notes_path)
        self.length_dict = length_dict
        self.allele_dict = {}
        self.profile_data = {}
        self.allele_links = {}
        self.aa_profile_file = os.path.join(self.report_path, 'aa_full_profile.txt')
        self.gene_names = set()
        self.gene_name_file = os.path.join(self.report_path, 'gene_names.txt')
        self.aa_profile_data = {}
        self.profile_matches = {}
        self.aa_nt_profile_link_file = os.path.join(
            self.report_path,
            'aa_nt_profile_links.tsv'
        )


def cli():
    """
    Collect the arguments, create an object, and run the script
    """
    # Parser for arguments
    parser = ArgumentParser(
        description='Translate allele files in nucleotide format to amino acid. '
                    'Remove duplicates. Keep notes.'
    )
    parser.add_argument(
        '-p', '--path',
        required=True,
        help='Specify path containing allele files.'
    )
    parser.add_argument(
        '--profile',
        action='store_true',
        help='Optionally parse the nucleic acid profile, and create the corresponding reduced '
             'amino acid profile. The profile must be named profile.txt and be located in nt_profile folder in the path'
    )
    parser.add_argument(
        '--report_path',
        default=os.path.join(os.getcwd(), 'aa_profile'),
        help='Optionally provide the name (and path, if desired) of the folder into which the amino acid profile '
             'and related files are to be written. Default folder is "aa_profile" in your current working directory'
    )
    parser.add_argument(
        '--translated_path',
        default=os.path.join(os.getcwd(), 'aa_alleles'),
        help='Optionally provide the name (and path, if desired) of the folder into which the amino acid alleles '
             'and notes are to be written. Default folder is "aa_alleles" in your current working directory'
    )
    # Get the arguments into an object
    arguments = parser.parse_args()
    SetupLogging(debug=True)
    translate = Translate(
        path=arguments.path,
        profile=arguments.profile,
        report_path=arguments.report_path,
        translated_path=arguments.translated_path
    )
    translate.main()
    logging.info('Allele translation complete!')
    # Prevent the arguments being printed to the console (they are returned in order for the tests to work)
    sys.stderr = open(os.devnull, 'w', encoding='utf-8')
    return arguments


if __name__ == '__main__':
    cli()



#!/usr/bin/env python3

"""
Perform allele profiling
"""

# Standard imports
from csv import DictReader
from glob import glob
import logging
import json
import os

# Third party imports
from Bio.Seq import Seq
from Bio import SeqIO


# Local imports
from allele_tools.methods import \
     extract_novel_alleles
from olctools.accessoryFunctions.accessoryFunctions import combinetargets
from genemethods.geneseekr.geneseekr import GeneSeekr


__author__ = 'adamkoziol'


def allele_prep(allele_path, gene_names, combined_targets, amino_acid):
    """
    Create a 'combinedtargets.fasta' file
    :param allele_path: Name and absolute path to folder containing allele database files
    :param gene_names: List of all gene names in the analysis
    :param combined_targets: String of absolute path to the combinedtargets.fasta file
    :param amino_acid: Boolean of whether the query sequences are amino acid or nucleotide
    :return: records: Dictionary of allele_header:allele_sequence
    :return: gene_names: List of all gene names extracted from allele headers
    :return: data: String of header including all gene names. To be used in creating final report
    """
    logging.info('Reading allele files')
    custom_targets = os.path.join(allele_path, 'custom.tfa')
    combined_targets_db = glob(os.path.join(allele_path, 'combinedtargets*'))
    records = {}
    # Clear out any previously created combinedtargets files (and associated BLAST databases)
    clear_alleles(combined_targets_db=combined_targets_db,
                  custom_targets=custom_targets)
    alleles = glob(os.path.join(allele_path, '*.*fa*'))
    # If the dictionary hasn't already been populated by a previous iteration, remove the path and
    # the extension from each of the allele files to determine gene names
    if not gene_names:
        for allele in alleles:
            gene_names.append(os.path.splitext(os.path.basename(allele))[0].replace('_alleles', ''))
    # Populate the header string for the final report
    genes = '\t'.join(sorted(gene_names))
    data = f'ST\t{genes}\n'
    # Create the combinedtargets file
    if not os.path.isfile(combined_targets):
        if amino_acid:
            combinetargets(
                targets=alleles,
                targetpath=allele_path,
                mol_type='prot'
            )
        else:
            combinetargets(
                targets=alleles,
                targetpath=allele_path
            )
    # Create BLAST databases
    GeneSeekr.makeblastdb(
        fasta=combined_targets,
        program='blastn' if not amino_acid else 'blastp'
    )
    # Load all the FASTA records from the combinedtargets file
    for allele_file in alleles:
        for record in SeqIO.parse(allele_file, 'fasta'):
            records[record.id] = str(record.seq)
    return records, gene_names, data


def clear_alleles(combined_targets_db, custom_targets):
    """
    Remove any combinedtargets.fasta or custom.tfa files present in the allele path
    :param combined_targets_db: List of combinedtargets files, including BLAST database files
    :param custom_targets: Name and absolute path to custom.tfa target file
    :return:
    """
    # Iterate through all the combinedtargets files
    for combined_target in combined_targets_db:
        # Remove each file
        os.remove(combined_target)
    # Attempt to remove the targets.tfa file
    try:
        os.remove(custom_targets)
    except FileNotFoundError:
        pass


def read_profile(profile_file):
    """
    Load any previously created profiles for this analysis
    :param profile_file: Name and absolute path to a profile file
    """
    # Initialise the dictionary
    profile_data = {}
    # Only load the profile file if it exists
    if os.path.isfile(profile_file):
        logging.info('Extracting profiles from profile file')
        # Open an Excel-formatted sequence profile file as a dictionary
        profile = DictReader(open(profile_file, encoding='utf-8'), dialect='excel-tab')
        # Iterate through the rows
        for row in profile:
            # Populate the profile dictionary with profile number: {gene: allele}.
            allele_comprehension = {gene: allele for gene, allele in row.items() if gene != 'ST'}
            # Extract the sequence type number from the first field name
            seq_type = row[profile.fieldnames[0]]
            # Update the profile data dictionary
            profile_data[seq_type] = allele_comprehension
    return profile_data


def parseable_blast_outputs(runmetadata, fieldnames, extended_fieldnames, records, cutoff=90):
    """
    Add a header to the BLAST report, so that it is easier to figure out what is in each column
    :param runmetadata: Metadata object containing a list of all metadata objects
    :param fieldnames: String of all the field names in the BLAST report
    :param extended_fieldnames: String of the BLAST field names plus the calculated percent identity
    :param records: Dictionary of allele name: allele sequence
    :param cutoff: Integer of the minimum percent identity between query and subject sequence. Default is 90
    """
    logging.info('Adding headers to BLAST outputs')
    for sample in runmetadata.samples:
        data = []
        # Load the first line of the report
        with open(sample.alleles.blast_report, 'r', encoding='utf-8') as report:
            header_line = report.readline().strip()
        # Split the header on tabs
        header_list = header_line.split('\t')
        # Check to see if the header has already been added. Skip this step if it has been added.
        if header_list[0] != fieldnames[0]:
            with open(sample.alleles.blast_report, 'r', encoding='utf-8') as report:
                header = list(report.readline().split('\t'))
            if len(header) == len(fieldnames):
                current_fieldnames = fieldnames
            else:
                current_fieldnames = extended_fieldnames
            blastdict = DictReader(open(sample.alleles.blast_report, encoding='utf-8'),
                                   fieldnames=current_fieldnames,
                                   dialect='excel-tab'
                                   )
            # Go through each BLAST result
            for row in blastdict:
                # Calculate the percent identity and extract the bit score from the row
                # Percent identity is the (length of the alignment - num mismatches) /
                # total query length
                percent_identity = float(
                    '{:0.2f}'.format((float(row['identical']) - float(row['gaps'])) /
                                     len(records[row['subject_id']]) * 100)
                )
                # Filter the results based on the cutoff value
                if percent_identity < cutoff:
                    continue
                # Create a percent match entry based on the calculated percent identity match
                row['percent_match'] = str(percent_identity)
                # Add the updated row to the list
                data.append(row)

            # Overwrite the original BLAST outputs to include headers, and the percent match
            with open(sample.alleles.blast_report, 'w', encoding='utf-8') as updated_report:
                # Add the header
                headers = '\t'.join(extended_fieldnames)
                updated_report.write(f'{headers}\n')
                # Add the results
                for row in data:
                    for header in extended_fieldnames:
                        # Write the value from the row with the header as the key
                        try:
                            updated_report.write('{value}\t'.format(value=row[header]))
                        except KeyError:
                            # noinspection PyTypeChecker
                            updated_report.write('{value}\t'.format(value=''.join(row[None])))
                    # Add a newline for each result
                    updated_report.write('\n')


def parse_results(runmetadata, fieldnames, extended_fieldnames, amino_acid, genome_query=False):
    """
    Parse the BLAST results, and populate GenObjects
    :param runmetadata: Metadata object containing list of all metadata objects
    :param fieldnames: String of all the field names in the BLAST report
    :param extended_fieldnames: String of the BLAST field names plus the calculated percent identity
    :param amino_acid: Variable on whether targets are protein
    :param genome_query: Boolean of whether the query is a genome, and the subject is the allele
    :return: Updated runmetadata
    """
    logging.info('Parsing BLAST outputs')
    for sample in runmetadata.samples:
        # Initialise GenObjects as required
        sample.alleles.blastlist = []
        sample.alleles.targetsequence = {}
        # Open the sequence profile file as a dictionary
        blastdict = DictReader(open(sample.alleles.blast_report, encoding='utf-8'),
                               fieldnames=extended_fieldnames,
                               dialect='excel-tab')
        resultdict = {}
        # Go through each BLAST result
        for row in blastdict:
            # Ignore the headers
            if row['query_id'].startswith(fieldnames[0]):
                pass
            else:
                target_id = row['query_id'] if not genome_query else row['subject_id']
                target_start = row['subject_start']
                target_end = row['subject_end']
                target_seq = row['query_sequence']
                # Remove unwanted pipes added to the name
                target = target_id.lstrip('gb|').rstrip('|') if '|' in target_id else \
                    target_id
                # If the percent identity is equal to the cutoff
                if float(row['percent_match']) == 100:
                    # Append the hit dictionary to the list
                    sample.alleles.blastlist.append(row)
                    # Update the dictionary with the target and percent identity
                    resultdict.update({target: row['percent_match']})
                    # Determine if the orientation of the sequence is reversed compared to
                    # the reference sequence
                    if int(target_end) < int(target_start) and not amino_acid:
                        # Create a sequence object using BioPython
                        seq = Seq(target_seq)
                        # Calculate the reverse complement of the sequence
                        querysequence = str(seq.reverse_complement())
                    # If the sequence is not reversed, use the sequence as it is in the output
                    else:
                        querysequence = target_seq
                    # Add the sequence in the correct orientation to the sample
                    try:
                        sample.alleles.targetsequence[target].append(querysequence)
                    except (AttributeError, KeyError):
                        sample.alleles.targetsequence[target] = []
                        sample.alleles.targetsequence[target].append(querysequence)
            # Add the percent identity to the object
            sample.alleles.blastresults = resultdict
        # Populate missing results with 'NA' values
        if len(resultdict) == 0:
            sample.alleles.blastresults = 'NA'
    return runmetadata


def profile_alleles(runmetadata, profile_dict, profile_set, records, amino_acid=False,
                    novel_alleles=False, genome_query=False, allele_path=None, report_path=None,
                    cutoff=75):
    """
    Create the gene:allele profile from the BLAST outputs from each sample
    :param runmetadata: Metadata object containing a list of all metadata objects
    :param profile_dict: Dictionary to store gene:allele profile for each sample
    :param profile_set: List of all unique profiles
    :param records: List of all gene names
    :param novel_alleles: Boolean of whether novel alleles should be extracted from BLAST hit if
    there is no 100% match
    :param genome_query: Boolean of whether the query is a genome, and the subject is the allele
    :param amino_acid: Variable on whether targets are protein
    :param allele_path: Name and absolute path to folder containing allele database files
    :param report_path: Name and absolute path to folder in which reports are to be created
    :param cutoff: Integer of the minimum percent identity between query and subject sequence. Default is 75
    :return: Updated profile_dict and profile_set
    """
    logging.info('Determining allele profiles')
    # Iterate through all the samples
    for sample in runmetadata.samples:
        # Initialise a dictionary to store the profile information for each samples
        profile_dict[sample.name] = {}
        # Initialise a dictionary to store the gene:allele combinations for each sample
        allele_comprehension = {}
        # Each gene in the analysis is stored in the list of genes created in allele_finder
        for gene in records:
            # Create a variable to track whether the current gene is present in the current sample
            present = False
            # Iterate through all the BLAST outputs for the sample
            for allele in sample.alleles.blastresults:
                # If the gene name is present as a substring of the allele e.g. adk in adk_1,
                # then the gene is present in the BLAST outputs
                if gene in allele:
                    # Strip off the allele number from the allele e.g. adk_1 yields 1
                    allele_id = allele.split('_')[-1]
                    # Update the dictionary with the new gene: allele number for the sample
                    allele_comprehension.update({gene: allele_id})
                    # Update the gene presence variable
                    present = True
            # If, after iterating through all the BLAST outputs, the gene is not present in the
            # sample, update the gene: allele to reflect this absence
            if not present:
                if novel_alleles:
                    sample, novel_allele, query_sequence = extract_novel_alleles(
                        sample=sample,
                        gene=gene,
                        genome_query=genome_query,
                        amino_acid=amino_acid,
                        allele_path=allele_path,
                        report_path=report_path,
                        cutoff=cutoff,
                    )
                    try:
                        sample.alleles.targetsequence[novel_allele].append(query_sequence)
                    except KeyError:
                        sample.alleles.targetsequence[novel_allele] = [query_sequence]
                    if novel_allele:
                        allele_comprehension.update({gene: novel_allele.split('_')[-1]})
                    else:
                        allele_comprehension.update({gene: '0'})
                else:
                    # Set missing alleles to 'ND'
                    allele_comprehension.update({gene: '0'})
        # In order to hash the dictionary, use JSON, with sorted keys to freeze it
        frozen_allele_comprehension = json.dumps(allele_comprehension, sort_keys=True)
        # Update the dictionary of profiles with the hash of the frozen dictionary: list of
        # samples with that hash
        if hash(frozen_allele_comprehension) not in profile_dict:
            profile_dict[hash(frozen_allele_comprehension)] = [sample.name]
        else:
            profile_dict[hash(frozen_allele_comprehension)].append(sample.name)
        # Add the 'regular' dictionary to the list of all profiles as required
        if allele_comprehension not in profile_set:
            profile_set.append(allele_comprehension)
    return profile_dict, profile_set


def match_profile(profile_data, profile_dict, profile_matches):
    """
    Match current profiles to any previously created profiles
    :param profile_data: Dictionary of seq_type: {gene name:allele ID}
    :param profile_dict: Dictionary of gene:allele profile for each sample
    :param profile_matches: Dictionary of seq_type: matching profile
    :return: profile_matches: Updated dictionary of seq_type: matching profiles
    """
    # If the profile_data dictionary was not populated in the read_profiles methods,
    # there is nothing to match
    if profile_data:
        logging.info('Matching new profiles against profile file')
        # Extract the sequence type and allele dictionary from the profile file
        for seq_type, allele_comprehension in profile_data.items():
            # Freeze the allele comprehension as above
            frozen_allele_comprehension = json.dumps(allele_comprehension, sort_keys=True)
            try:
                # Extract the samples that match this profile
                matches = profile_dict[hash(frozen_allele_comprehension)]
                # Update the dictionary with the matching samples
                profile_matches[seq_type] = matches
            # The profile will not necessarily match any of the profiles found in the analysis
            except KeyError:
                pass
    return profile_matches


def create_profile(profile_data, profile_set, new_profiles, profile_dict, profile_matches):
    """
    Create new profiles for novel profiles as required
    :param profile_data: Dictionary of seq_type: {gene name:allele ID}
    :param profile_set: List of all unique profiles
    :param new_profiles: List of novel profiles
    :param profile_dict: Dictionary of gene:allele profile for each sample
    :param profile_matches: Dictionary of seq_type: matching profile
    :return: profile_matches: Updated dictionary of seq_type: matching profiles
    :return: profile_data: Updated dictionary of seq_type: {gene name: allele ID}
    :return: new_profiles: Updated list of novel profiles
    """
    # Initialise the sequence type to be 1
    seq_type = 1
    # If the profile_data dictionary exists, set the sequence type to be the last of the entries
    # in the dictionary plus one, as that corresponds to the next sequence type
    if profile_data:
        # seq_type = len(profile_data) + 1
        seq_type = sorted(int(st) for st in profile_data.keys())[-1] + 1
    # Initialise a list to store the matched samples
    matched = []
    # Iterate through all the profiles in the analysis
    for allele_comprehension in profile_set:
        # Ensure that the allele comprehension (profile) is not already in the profile file
        if allele_comprehension not in [profiled_alleles for st, profiled_alleles in profile_data.items()]:
            # Add the new profile to the list of new profiles
            alleles = '\t'.join(
                allele_num.split('_')[-1] for gene, allele_num in sorted(
                    allele_comprehension.items()
                )
            )
            new_profiles.append(
                f'{seq_type}\t{alleles.rstrip()}'
            )
            # Freeze the comprehension in order to be used as the key in the profile dictionary
            frozen_allele_comprehension = json.dumps(allele_comprehension, sort_keys=True)
            matches = profile_dict[hash(frozen_allele_comprehension)]
            # Check to see if this sequence type hasn't already been found in the current analysis
            if matches not in matched:
                # Update the dictionary with the new sequence type: list of samples
                profile_matches[seq_type] = matches
                profile_data[seq_type] = allele_comprehension
                # Add the matches to the list of matches
                matched.append(matches)
            # Increment the sequence type number of the next entry
            seq_type += 1
    return profile_matches, profile_data, new_profiles


def sequence_typer(profile_report, data, runmetadata, profile_matches, profile_data,
                   update=False, amino_acid=False):
    """
    Perform the final sequence typing, and create the report
    :param profile_report: String of absolute path to report file
    :param data: String of header including all gene names. To be used in creating final report
    :param runmetadata: Metadata object containing a list of all metadata objects
    :param profile_matches: Dictionary of seq_type: matching profile
    :param profile_data: Dictionary of seq_type: {gene name:allele ID}
    :param update: Boolean of whether the report is to be created or updated. Default is False (created)
    :param amino_acid: Boolean of whether the query sequences are amino acid. Default is False
    """
    # Open the report
    mode = 'w' if not update else 'a+'
    if not update:
        # Initialise the header with an extra 'Sample' column plus the comma-separate list
        # of gene names
        data = 'Sample\t' + data
    else:
        if not os.path.isfile(profile_report):
            # Initialise the header with an extra 'Sample' column plus the comma-separate list
            # of gene names
            data = 'Sample\t' + data
        else:
            data = str()
    with open(profile_report, mode, encoding='utf-8') as report:
        for sample in runmetadata.samples:
            # Iterate through all the matches to sequence profiles in the analysis
            for seq_type, sample_names in profile_matches.items():
                # Check if the sample name is in the list of samples names with the current
                # sequence type
                if sample.name in sample_names:
                    # Add the sample name, sequence type, and all the allele numbers to the
                    # report string
                    complement = '\t'.join(allele_num.split('_')[-1] for gene, allele_num in sorted(
                        profile_data[seq_type].items())
                    )
                    data += f'{sample.name}\t{seq_type}\t{complement.rstrip()}\n'
                    # Update the appropriate GenObject based on the current molecule
                    # (DNA or amino acid)
                    if not amino_acid:
                        sample.alleles.nt_st = seq_type
                        sample.alleles.nt_profile = profile_data[seq_type]
                    else:
                        sample.alleles.aa_st = seq_type
                        sample.alleles.aa_profile = profile_data[seq_type]
        # Write the report
        report.write(data)
    return runmetadata


def append_profiles(new_profiles, profile_file, data, novel_profiles=False, profile_path=None, gene_names=None):
    """
    Add new profiles to the profile file
    :param new_profiles: List of all novel profiles in this analysis
    :param profile_file: Name and absolute path to a profile file
    :param data: String of header including all gene names. To be used in creating final report
    :param novel_profiles: Boolean of whether the novel_profiles.txt file is to be populated. Default is False
    :param profile_path: String of absolute path of folder in which profiles are located
    :param gene_names: List of all genes in the analysis
    """
    # Only try to add new profiles if there are new profiles in the analysis
    if new_profiles:
        # Initialise the string to store the new profile
        new_data = str()
        # If the profile file does not exist, add the string of 'ST', comma-separated gene names
        # to the headers
        if not os.path.isfile(profile_file):
            new_data = data
        # Iterate through all the new profiles, and add them to the new profile string
        for profile in new_profiles:
            new_data += f'{profile}\n'
        # Open the report with a+ to either create, or append the profile string to it
        with open(profile_file, 'a+', encoding='utf-8') as profile:
            profile.write(new_data)
        if novel_profiles:
            novel_profile_file = os.path.join(profile_path, 'novel_profiles.txt')
            if not os.path.isfile(novel_profile_file):
                with open(novel_profile_file, 'w', encoding='utf-8') as novel:
                    novel.write('ST\t{names}\n'.format(names='\t'.join(gene_names)))
            with open(novel_profile_file, 'a+', encoding='utf-8') as novel:
                novel.write(new_data)



#!/usr/bin/env python

"""
Run allele finding specifically for the co-located subunits of STEC genes
"""

# Standard imports
from argparse import (
    ArgumentParser,
    RawTextHelpFormatter
)
from glob import glob
import logging
import multiprocessing
import sys
import os

# Third party inputs
from olctools.accessoryFunctions.accessoryFunctions import (
    make_path,
    MetadataObject,
)

# Local imports
from allele_tools.allele_profiler import (
    allele_prep,
    parseable_blast_outputs,
    read_profile
)
from allele_tools.allele_translate_reduce import Translate
from allele_tools.profile_reduce import ProfileReduce
from allele_tools.version import __version__
from allele_tools.methods import (
    analyse_aa_alleles,
    blast_alleles,
    common_allele_find_errors,
    concatenate_alleles,
    create_aa_allele_comprehension,
    create_frozen_allele_comprehension,
    create_gene_names,
    create_nt_allele_comprehension,
    create_stec_report,
    error_print,
    load_alleles,
    match_profile,
    parse_aa_blast,
    parse_colocated_results,
    pathfinder,
    profile_allele_check,
    query_prep,
    report_aa_alleles,
    setup_arguments,
    split_alleles,
    write_concatenated_sequences
)

__author__ = 'adamkoziol'


class STEC:
    """
    Perform the STEC-specific allele finding
    """

    def main(self):
        """
        Run the appropriate methods in the correct order
        """
        # Create metadata objects for all files in the query folder
        self.runmetadata = query_prep(
            query_path=self.query_path,
            runmetadata=self.runmetadata
        )
        for sample in self.runmetadata.samples:
            logging.debug('Processing sample %s', sample.name)
            records, gene_names, self.data = \
                allele_prep(
                    allele_path=self.nt_allele_path,
                    gene_names=self.gene_names,
                    combined_targets=self.combined_targets,
                    amino_acid=self.amino_acid
                )
            gene_names = sorted(gene_names)
            logging.info('Loading profile')
            nt_profile_data = read_profile(profile_file=self.nt_profile_file)
            aa_profile_data = read_profile(profile_file=self.aa_profile_file)
            # Extract the sequence type and allele dictionary from the profile file
            blast_alleles(
                runmetadata=sample,
                amino_acid=self.amino_acid,
                combined_targets=self.combined_targets,
                cpus=self.cpus,
                outfmt=self.outfmt
            )
            # Add the header to the BLAST outputs, and filter results based on cutoff
            parseable_blast_outputs(
                runmetadata=sample,
                fieldnames=self.fieldnames,
                extended_fieldnames=self.extended_fieldnames,
                records=records,
                cutoff=90
            )
            # Parse the BLAST results
            sample, notes = parse_colocated_results(
                runmetadata=sample,
                fieldnames=self.fieldnames,
                extended_fieldnames=self.extended_fieldnames,
                amino_acid=self.amino_acid,
                gene_names=gene_names,
                nt_allele_path=self.nt_allele_path,
                aa_allele_path=self.aa_allele_path,
                report_path=self.report_path
            )
            # Create nucleotide allele comprehensions from the BLAST outputs
            nt_allele_comprehension = create_nt_allele_comprehension(
                runmetadata=sample,
                gene_names=gene_names
            )
            # Create an amino acid allele comprehensions from the translated BLAST outputs
            aa_allele_comprehension = create_aa_allele_comprehension(
                runmetadata=sample,
                gene_names=gene_names,
            )
            # Freeze the nucleotide allele comprehension
            nt_frozen_allele_comprehension = create_frozen_allele_comprehension(
                allele_comprehension=nt_allele_comprehension
            )
            # Freeze the amino acid allele comprehension
            aa_frozen_allele_comprehension = create_frozen_allele_comprehension(
                allele_comprehension=aa_allele_comprehension
            )
            # Find nucleotide profile matches
            nt_profile_matches, nt_frozen_profiles = match_profile(
                profile_data=nt_profile_data,
                frozen_allele_comprehension=nt_frozen_allele_comprehension,
                report_path=self.report_path,
                profile_file=self.nt_profile_file,
                genes=gene_names,
                allele_comprehension=nt_allele_comprehension,
                molecule='nt'
            )
            # Find amino acid profile matches
            aa_profile_matches, aa_frozen_profiles = match_profile(
                profile_data=aa_profile_data,
                frozen_allele_comprehension=aa_frozen_allele_comprehension,
                report_path=self.report_path,
                profile_file=self.aa_profile_file,
                genes=gene_names,
                allele_comprehension=aa_allele_comprehension,
                molecule='aa'
            )
            # Create the STEC-specific report
            create_stec_report(
                runmetadata=sample,
                nt_profile_matches=nt_profile_matches,
                nt_alleles=nt_allele_comprehension,
                aa_profile_matches=aa_profile_matches,
                aa_alleles=aa_allele_comprehension,
                report_file=self.report_file,
                gene_names=gene_names,
                aa_profile_path=self.aa_profile_path,
                notes=notes
            )

    def __init__(self, allele_path, aa_allele_path, profile_file, aa_profile_file, query_path, report_path,
                 amino_acid=False):

        self.nt_allele_path = pathfinder(path=allele_path)
        self.aa_allele_path = pathfinder(path=aa_allele_path)
        self.nt_profile_file = pathfinder(path=profile_file)
        self.aa_profile_file = pathfinder(path=aa_profile_file)
        self.profile_path = os.path.dirname(self.nt_profile_file)
        self.aa_profile_path = os.path.dirname(self.aa_profile_file)
        self.query_path = pathfinder(path=query_path)
        self.report_path = pathfinder(path=report_path)
        self.aa_report_path = self.report_path
        make_path(inpath=self.report_path)
        novel_alleles = glob(os.path.join(self.report_path, '*.fasta'))
        for novel_allele in novel_alleles:
            os.remove(novel_allele)

        self.amino_acid = amino_acid
        if not self.amino_acid:
            self.combined_targets = os.path.join(self.nt_allele_path, 'combinedtargets.fasta')
        else:
            self.combined_targets = os.path.join(self.aa_allele_path, 'combinedtargets.fasta')
        self.gene_names = []
        self.runmetadata = MetadataObject()
        self.runmetadata.samples = []
        self.cpus = multiprocessing.cpu_count() - 1
        self.profile_report = os.path.join(self.report_path, 'nt_profiles.tsv')
        self.aa_profile_report = os.path.join(self.aa_report_path, 'aa_profiles.tsv')
        try:
            os.remove(self.profile_report)
        except FileNotFoundError:
            pass

        self.report_file = os.path.join(self.report_path, 'stec_report.tsv')
        reports = glob(os.path.join(self.report_path, '*.tsv'))
        for report in reports:
            os.remove(report)
        # Fields used for custom outfmt 6 BLAST output:
        self.fieldnames = [
            'query_id',
            'subject_id',
            'identical',
            'mismatches',
            'gaps',
            'evalue',
            'bit_score',
            'query_length',
            'subject_length',
            'alignment_length',
            'query_start',
            'query_end',
            'subject_start',
            'subject_end',
            'query_sequence',
            'subject_sequence'
        ]
        self.extended_fieldnames = self.fieldnames.copy()
        self.extended_fieldnames.insert(14, 'percent_match')
        self.outfmt = '6 qseqid sseqid nident mismatch gaps evalue bitscore qlen slen length ' \
                      'qstart qend sstart send qseq sseq'
        # A string of the header to use for formatting the profile file, and the report headers
        self.data = str()
        self.aa_allele_dict = {}
        self.aa_nt_allele_link_dict = {}


class AASTEC:

    def main(self):
        """
        Run the necessary methods in the correct order
        """
        # Create metadata objects for all files in the query folder
        self.runmetadata = query_prep(
            query_path=self.query_path,
            runmetadata=self.runmetadata,
            clear_report=True
        )

        for sample in self.runmetadata.samples:
            logging.debug('Processing sample %s', sample.name)
            notes = []
            records, gene_names, self.data = \
                allele_prep(
                    allele_path=self.aa_allele_path,
                    gene_names=self.gene_names,
                    combined_targets=self.combined_targets,
                    amino_acid=self.amino_acid
                )
            gene_names = sorted(gene_names)
            logging.info('Loading profile')
            if not os.path.isfile(sample.alleles.blast_report):
                #
                blast_alleles(
                    runmetadata=sample,
                    amino_acid=self.amino_acid,
                    combined_targets=self.combined_targets,
                    cpus=self.cpus,
                    outfmt=self.outfmt
                )
                # Add headers to the BLAST outputs, and filter based on cutoff value
                parseable_blast_outputs(
                    runmetadata=sample,
                    fieldnames=self.fieldnames,
                    extended_fieldnames=self.extended_fieldnames,
                    records=records,
                    cutoff=self.cutoff
                )
            # Parse the amino acid BLAST results
            sample, filtered, notes = parse_aa_blast(
                runmetadata=sample,
                extended_fieldnames=self.extended_fieldnames,
                fieldnames=self.fieldnames,
                gene_names=gene_names,
                notes=notes,
                aa_allele_path=self.aa_allele_path,
                report_path=self.report_path,
                cutoff=self.cutoff
            )
            # Perform content/length checks of the supplied alleles
            sample, notes = analyse_aa_alleles(
                runmetadata=sample,
                gene_names=gene_names,
                notes=notes)
            # Create an amino acid query report
            report_aa_alleles(
                runmetadata=sample,
                report_file=self.report_file,
                notes=notes
            )

    def __init__(self, aa_allele_path, query_path, report_path, cutoff, amino_acid=True):

        self.aa_allele_path = pathfinder(path=aa_allele_path)
        self.query_path = pathfinder(path=query_path)
        self.report_path = pathfinder(path=report_path)
        self.aa_report_path = self.report_path
        make_path(inpath=self.report_path)
        novel_alleles = glob(os.path.join(self.report_path, '*.fasta'))
        for novel_allele in novel_alleles:
            os.remove(novel_allele)
        self.cutoff = cutoff
        self.amino_acid = amino_acid
        self.combined_targets = os.path.join(self.aa_allele_path, 'combinedtargets.fasta')
        self.gene_names = []
        self.runmetadata = MetadataObject()
        self.runmetadata.samples = []
        self.cpus = multiprocessing.cpu_count() - 1
        self.report_file = os.path.join(self.report_path, 'allele_report.tsv')
        try:
            os.remove(self.report_file)
        except FileNotFoundError:
            pass
        # Fields used for custom outfmt 6 BLAST output:
        self.fieldnames = [
            'query_id',
            'subject_id',
            'identical',
            'mismatches',
            'gaps',
            'evalue',
            'bit_score',
            'query_length',
            'subject_length',
            'alignment_length',
            'query_start',
            'query_end',
            'subject_start',
            'subject_end',
            'query_sequence',
            'subject_sequence'
        ]
        self.extended_fieldnames = self.fieldnames.copy()
        self.extended_fieldnames.insert(14, 'percent_match')
        self.outfmt = '6 qseqid sseqid nident mismatch gaps evalue bitscore qlen slen length ' \
                      'qstart qend sstart send qseq sseq'
        # A string of the header to use for formatting the profile file, and the report headers
        self.data = str()


class AlleleConcatenate:

    """
    Concatenate stx subunits. Read in profile files. Load alleles. Concatenate alleles with appropriate linker
    """

    def main(self):
        """
        Run the necessary methods for AlleleConcatenate
        """
        self.gene, self.nt_alleles = load_alleles(
            allele_path=self.nt_allele_path,
            allele_order=self.allele_order
        )
        self.gene, self.aa_alleles = load_alleles(
            allele_path=self.aa_allele_path,
            allele_order=self.allele_order
        )
        logging.info('Concatenating allele sequences')
        self.concatenated_nt_seq = concatenate_alleles(
            profile_data=self.nt_profile_data,
            allele_dict=self.nt_alleles,
            allele_order=self.allele_order,
            stx_gene=self.gene,
            linker_length_dict=self.linker_length_dict,
            molecule='nt'
        )
        self.concatenated_aa_seq = concatenate_alleles(
            profile_data=self.aa_profile_data,
            allele_dict=self.aa_alleles,
            allele_order=self.allele_order,
            stx_gene=self.gene,
            linker_length_dict=self.linker_length_dict,
            molecule='aa'
        )
        logging.info('Writing concatenated allele sequences to file')
        write_concatenated_sequences(
            concatenated_sequences=self.concatenated_nt_seq,
            concatenate_path=self.concatenate_path,
            file_name=self.gene_allele[self.gene],
            molecule='nt'
        )
        write_concatenated_sequences(
            concatenated_sequences=self.concatenated_aa_seq,
            concatenate_path=self.concatenate_path,
            file_name=self.gene_allele[self.gene],
            molecule='aa'
        )

    def __init__(
            self,
            nt_allele_path,
            aa_allele_path,
            nt_profile_file,
            aa_profile_file,
            concatenate_path):
        self.nt_allele_path = pathfinder(path=nt_allele_path)
        self.aa_allele_path = pathfinder(path=aa_allele_path)
        self.nt_profile_file = pathfinder(path=nt_profile_file)
        self.aa_profile_file = pathfinder(path=aa_profile_file)
        self.nt_profile_path = os.path.dirname(self.nt_profile_file)
        self.aa_profile_path = os.path.dirname(self.aa_profile_file)
        self.concatenate_path = pathfinder(path=concatenate_path)
        self.linker_length_dict = {
            'stx1': 9,
            'stx2': 12,
        }
        # Set the appropriate order for the genes in the report (stx1 genes are not in numerical order)
        self.allele_order = {
            'stx1': ['ECs2974', 'ECs2973'],
            'stx2': ['ECs1205', 'ECs1206']
        }
        self.gene_allele = {
            'stx1': 'ECs2974_ECs2973',
            'stx2': 'ECs1205_ECs1206'
        }
        self.nt_profile_data = read_profile(profile_file=self.nt_profile_file)
        self.aa_profile_data = read_profile(profile_file=self.aa_profile_file)
        self.gene = str()
        self.nt_alleles = {}
        self.aa_alleles = {}
        self.concatenated_nt_seq = []
        self.concatenated_aa_seq = []


def profile_reduce(args):
    """
    Reduce the Enterobase profile to only the genes of interest
    :param args: type ArgumentParser arguments
    """
    # Create the gene names file if it doesn't exist or is empty
    genes_path = os.path.dirname(args.gene_names)
    genes_file = os.path.basename(args.gene_names)
    logging.info(genes_path)
    logging.info(genes_file)
    if not os.path.isfile(args.gene_names):
        # Ensure that the path exists
        if not os.path.isdir(genes_path):
            logging.error(
                'Could not locate the supplied path, %s, for the gene file. Please ensure that it'
                ' exists, and that it either contains the gene file, or files with '
                '.fasta extensions',
                genes_path
            )
            raise SystemExit
        logging.warning(
            'Could not locate the supplied gene file: %s. Will now attempt to create it in '
            'directory %s',
            args.gene_names, genes_path
            )
        # Attempt to create the file
        create_gene_names(
            path=genes_path,
            name=genes_file
        )
    else:
        # Ensure that the file isn't empty
        if os.stat(args.gene_names).st_size == 0:
            logging.warning(
                'The supplied gene file, %s, is empty. Will now attempt to populate it in '
                'directory %s',
                args.gene_names, genes_path
            )
            # Attempt to repopulate the file
            create_gene_names(
                path=genes_path,
                name=genes_file
            )
    logging.info(
        'Reducing profile file %s to include only the genes found in %s',
        args.profile_file, args.gene_names
        )
    # Create a ProfileReduce object
    profile_reduction = ProfileReduce(
        profile=args.profile_file,
        names=args.gene_names,
        output=args.output_folder
    )
    profile_reduction.main()
    logging.info('Profile reduction complete!')


def translate_reduce(args):
    """
    Translate nucleotide alleles to amino acid, and remove duplicates
    :param args: type ArgumentParser arguments
    """
    log_str = f'Translating and reducing alleles in: {args.allele_path}'
    if args.profile_file:
        log_str += f'. Parsing {args.profile_file} nucleotide profile to create corresponding,'\
                    ' reduced amino acid profile'
    logging.info(log_str)
    length_dict = {
        'ECs2973': 90,
        'ECs2974': 316,
        'ECs1205': 316,
        'ECs1206': 88
    }
    allele_translate_reduce = Translate(
        path=args.allele_path,
        profile=args.profile_file,
        report_path=args.report_path,
        translated_path=args.translated_path,
        length_dict=length_dict,
    )
    allele_translate_reduce.main()
    logging.info('Allele translation and reduction complete!')


def allele_find(args):
    """
    Perform allele discovery analyses
    :param args: type ArgumentParser arguments
    """
    log_str = f'Performing STEC allele discovery on sequences in {args.query_path} using ' \
        f'nucleotide alleles in {args.nt_alleles}, nucleotide profile in {args.nt_profile}, amino' \
        f' acid alleles in {args.aa_alleles}, and amino acid profile in {args.aa_profile}'
    logging.info(log_str)
    errors = []

    # Nucleotide allele checks
    if not os.path.isdir(args.nt_alleles):
        errors.append(f'Could not find supplied nucleotide allele folder: {args.nt_alleles}')
    else:
        if not glob(os.path.join(args.nt_alleles, '*.fasta')):
            errors.append(
                f'Could not locate sequence files in supplied nucleotide allele folder: {args.nt_alleles}'
            )
    # Find errors for amino acid and query checks
    errors = common_allele_find_errors(
        args=args,
        errors=errors,
        amino_acid=False
    )
    if not os.path.isfile(args.nt_profile):
        errors.append(f'Could not locate supplied nucleotide profile file: {args.nt_profile}')
    if errors:
        error_print(errors=errors)
    stec = STEC(
        allele_path=args.nt_alleles,
        aa_allele_path=args.aa_alleles,
        profile_file=args.nt_profile,
        aa_profile_file=args.aa_profile,
        query_path=args.query_path,
        report_path=args.report_path
    )
    stec.main()


def aa_allele_find(args):
    """
    Perform allele discovery analyses on amino acid query files
    :param args: type ArgumentParser arguments
    """
    log_str = f'Performing STEC allele discovery on amino acid sequences in {args.query_path} using amino' \
              f' acid alleles in {args.aa_alleles}'
    logging.info(log_str)
    errors = []
    # Find errors for amino acid and query checks
    errors = common_allele_find_errors(
        args=args,
        errors=errors,
        amino_acid=True
    )
    if errors:
        error_print(errors=errors)
    aa_stec = AASTEC(
        aa_allele_path=args.aa_alleles,
        query_path=args.query_path,
        report_path=args.report_path,
        cutoff=args.cutoff
    )
    aa_stec.main()


def allele_split(args):
    """
    Split files containing multiple alleles into individual files
    :param args: type ArgumentParser arguments
    """
    logging.info('Splitting allele files in %s into individual files', args.query_path)
    errors = []
    allele_files = glob(os.path.join(args.query_path, '*.fasta'))
    # Query checks
    if not os.path.isdir(args.query_path):
        errors.append(f'Could not find supplied nucleotide allele folder: {args.query_path}')
    else:
        if not allele_files:
            errors.append(
                f'Could not locate sequence files in supplied allele folder: {args.query_path}'
            )
    if errors:
        error_print(errors=errors)
    split_alleles(
        allele_files=allele_files,
        output_path=args.output_path
    )


def allele_concatenate(args):
    """
    Concatenate subunit files with linkers. Provide linkages between nucleotide and amino acid files
    :param args: type ArgumentParser arguments
    """
    logging.info('Concatenating allele subunits')
    errors = []
    # Determine if the profile file, the allele folder, and the alleles all exist
    errors = profile_allele_check(
        args=args,
        errors=errors
    )
    # If there were any errors with the supplied arguments, print them, and exit
    if errors:
        error_print(errors=errors)
    concatenate = AlleleConcatenate(
        nt_allele_path=args.nt_alleles,
        aa_allele_path=args.aa_alleles,
        nt_profile_file=args.nt_profile,
        aa_profile_file=args.aa_profile,
        concatenate_path=args.concatenate_path
    )
    concatenate.main()


def cli():
    """
    Collect the arguments, create an object, and run the script
    """
    # Parser for arguments
    parser = ArgumentParser(
        description='Determines STEC subunit profiles'
        )
    subparsers = parser.add_subparsers(title='Available analyses')
    # Create a parental parser from subparsers can inherit arguments
    parent_parser = ArgumentParser(add_help=False)
    # Add arguments common to all subparsers to the parent parser
    parent_parser.add_argument(
        '-version', '--version',
        action='version',
        version=f'%(prog)s commit {__version__}'
        )
    parent_parser.add_argument(
        '-v', '--verbosity',
        choices=['debug', 'info', 'warning', 'error', 'critical'],
        metavar='verbosity',
        default='info',
        help='Set the logging level. Options are debug, info, warning, error, and critical. '
             'Default is info.'
    )
    profile_reduce_subparser = subparsers.add_parser(
        parents=[parent_parser],
        name='profile_reduce',
        description='Reduce full wgMLST profile from Enterobase using genes of interest',
        formatter_class=RawTextHelpFormatter,
        help='Reduce full wgMLST profile from Enterobase using genes of interest'
    )
    profile_reduce_subparser.add_argument(
        '-p', '--profile_file',
        metavar='profile_file',
        default='profiles.list',
        help='Specify name and path of profile file. If not provided, the default '
             '"profiles.list" in the current working directory will be used'
        )
    profile_reduce_subparser.add_argument(
        '-g', '--gene_names',
        metavar='gene_names',
        default='genes.txt',
        help='Name and path of text file containing gene names to use to filter the profile file '
             '(one per line). If not provided, the default "genes.txt" in the current working '
             'directory will be used. If the file does not exist, the program will attempt to '
             'create a file using the .fasta files in the current working directory'
    )
    profile_reduce_subparser.add_argument(
        '-o', '--output_folder',
        metavar='output_folder',
        default='nt_profile',
        help='Name and path of folder into which the reduced profile and notes are to be placed. '
             'If not provided, the default "nt_profile" folder in the current working directory '
             'will be used'
    )
    profile_reduce_subparser.set_defaults(func=profile_reduce)
    # Create a subparser for allele translation and reduction
    allele_translate_reduce_subparser = subparsers.add_parser(
        parents=[parent_parser],
        name='allele_translate_reduce',
        description='Translate allele files in nucleotide format to amino acid. Remove duplicates. Keep notes',
        formatter_class=RawTextHelpFormatter,
        help='Translate allele files in nucleotide format to amino acid. '
             'Remove duplicates. Keep notes'
    )
    allele_translate_reduce_subparser.add_argument(
        '-a', '--allele_path',
        metavar='allele_path',
        default=os.path.join(os.getcwd(), 'nt_alleles'),
        help='Specify name and path of folder containing allele files. If not provided, the '
             'nt_alleles folder in the current working directory will be used by default'
    )
    allele_translate_reduce_subparser.add_argument(
        '-p', '--profile_file',
        metavar='profile_file',
        help='Optionally specify name and path of profile file. '
             'Parse the nucleic acid profile, and create the corresponding reduced amino acid profile'
        )
    allele_translate_reduce_subparser.add_argument(
        '-r', '--report_path',
        metavar='report_path',
        default=os.path.join(os.getcwd(), 'aa_profile'),
        help='Specify the name and path of the folder into which outputs are to be placed. If not '
              'provided, the aa_profile folder in the current working directory will be used'
    )
    allele_translate_reduce_subparser.add_argument(
        '-t', '--translated_path',
        metavar='translated_path',
        default=os.path.join(os.getcwd(), 'aa_alleles'),
        help='Specify the name and path of the folder into which alleles are to be placed. If not '
              'provided, the aa_alleles folder in the current working directory will be used'
    )
    allele_translate_reduce_subparser.set_defaults(func=translate_reduce)
    # Create a subparser for allele discovery
    allele_find_subparser = subparsers.add_parser(
        parents=[parent_parser],
        name='allele_find',
        description='Analyse sequences to determine allele complement. Update profiles and '
        'databases. Keep notes',
        formatter_class=RawTextHelpFormatter,
        help='Analyse sequences to determine allele complement. Update profiles and databases. '
        'Keep notes'
    )
    allele_find_subparser.add_argument(
        '--nt_profile',
        metavar='nt_profile',
        default=os.path.join(os.getcwd(), 'nt_profile', 'profile.txt'),
        help='Specify name and path of nucleotide profile file. If not provided, profile.txt in '
             'the nt_profile folder in the current working directory will be used by default'
    )
    allele_find_subparser.add_argument(
        '--aa_profile',
        metavar='aa_profile',
        default=os.path.join(os.getcwd(), 'aa_profile', 'profile.txt'),
        help='Specify name and path of amino acid profile file. If not provided, profile.txt in '
             'the aa_profile folder in the current working directory will be used by default'
    )
    allele_find_subparser.add_argument(
        '--nt_alleles',
        metavar='nt_alleles',
        default=os.path.join(os.getcwd(), 'nt_alleles'),
        help='Specify name and path of folder containing nucleotide alleles. If not provided, the '
             'nt_allele folder in the current working directory will be used by default'
    )
    allele_find_subparser.add_argument(
        '--aa_alleles',
        metavar='aa_alleles',
        default=os.path.join(os.getcwd(), 'aa_alleles'),
        help='Specify name and path of folder containing amino acid alleles. If not provided, the '
             'aa_allele folder in the current working directory will be used by default'
    )
    allele_find_subparser.add_argument(
        '-r', '--report_path',
        metavar='report_path',
        default=os.path.join(os.getcwd(), 'reports'),
        help='Specify name and path of folder into which reports are to be placed. If not '
             'provided, the reports folder in the current working directory will be used'
    )
    allele_find_subparser.add_argument(
        '-q', '--query_path',
        metavar='query_path',
        default=os.path.join(os.getcwd(), 'query'),
        help='Specify name and path of folder containing query files in FASTA format. '
             'If not provided, the query folder in the current working directory will be used'
    )
    allele_find_subparser.set_defaults(func=allele_find)
    # Create a subparser for allele discovery
    aa_allele_find_subparser = subparsers.add_parser(
        parents=[parent_parser],
        name='aa_allele_find',
        description='Analyse amino acid sequences to determine allele complement. Update profiles and '
                    'databases. Keep notes',
        formatter_class=RawTextHelpFormatter,
        help='Analyse amino acid sequences to determine allele complement. Update profiles and databases. '
             'Keep notes'
    )
    aa_allele_find_subparser.add_argument(
        '--aa_alleles',
        metavar='aa_alleles',
        default=os.path.join(os.getcwd(), 'aa_alleles'),
        help='Specify name and path of folder containing amino acid alleles. If not provided, the '
             'aa_allele folder in the current working directory will be used by default'
    )
    aa_allele_find_subparser.add_argument(
        '-r', '--report_path',
        metavar='report_path',
        default=os.path.join(os.getcwd(), 'reports'),
        help='Specify name and path of folder into which reports are to be placed. If not '
             'provided, the reports folder in the current working directory will be used'
    )
    aa_allele_find_subparser.add_argument(
        '-q', '--query_path',
        metavar='query_path',
        default=os.path.join(os.getcwd(), 'query'),
        help='Specify name and path of folder containing query files in FASTA format. '
             'If not provided, the query folder in the current working directory will be used'
    )
    aa_allele_find_subparser.add_argument(
        '-c', '--cutoff',
        metavar='cutoff',
        default=90,
        choices=[percent for percent in range(90, 101)],
        help='Specify the percent identity cutoff for matches. Allowed values are between 90 and 100. Default is 100'
    )
    aa_allele_find_subparser.set_defaults(func=aa_allele_find)
    # Create a subparser for splitting multi-FASTA files of alleles
    allele_split_subparser = subparsers.add_parser(
        parents=[parent_parser],
        name='allele_split',
        description='Split combined allele files into individual files',
        formatter_class=RawTextHelpFormatter,
        help='Split combined allele files into individual files'
    )
    allele_split_subparser.add_argument(
        '-q', '--query_path',
        metavar='query_path',
        default=os.path.join(os.getcwd(), 'query'),
        help='Specify name and path of folder containing query files in FASTA format. '
             'If not provided, the query folder in the current working directory will be used'
    )
    allele_split_subparser.add_argument(
        '-o', '--output_path',
        metavar='output_path',
        default=os.path.join(os.getcwd(), 'split_alleles'),
        help='Specify name and path of folder into which the split allele files are to be written. '
             'If not provided, the split_alleles folder in the current working directory will be used'
    )
    allele_split_subparser.set_defaults(func=allele_split)
    # Create a subparser for concatenating stx subunits
    allele_concatenate_subparser = subparsers.add_parser(
        parents=[parent_parser],
        name='allele_concatenate',
        description='Concatenate stx toxin subunit alleles with linkers',
        formatter_class=RawTextHelpFormatter,
        help='Concatenate stx toxin subunit alleles with linkers'
    )
    allele_concatenate_subparser.add_argument(
        '--nt_profile',
        metavar='nt_profile',
        default=os.path.join(os.getcwd(), 'nt_profile', 'profile.txt'),
        help='Specify name and path of nucleotide profile file. If not provided, profile.txt in '
             'the nt_profile folder in the current working directory will be used by default'
    )
    allele_concatenate_subparser.add_argument(
        '--aa_profile',
        metavar='aa_profile',
        default=os.path.join(os.getcwd(), 'aa_profile', 'profile.txt'),
        help='Specify name and path of amino acid profile file. If not provided, profile.txt in '
             'the aa_profile folder in the current working directory will be used by default'
    )
    allele_concatenate_subparser.add_argument(
        '--nt_alleles',
        metavar='nt_alleles',
        default=os.path.join(os.getcwd(), 'nt_alleles'),
        help='Specify name and path of folder containing nucleotide alleles. If not provided, the '
             'nt_allele folder in the current working directory will be used by default'
    )
    allele_concatenate_subparser.add_argument(
        '--aa_alleles',
        metavar='aa_alleles',
        default=os.path.join(os.getcwd(), 'aa_alleles'),
        help='Specify name and path of folder containing amino acid alleles. If not provided, the '
             'aa_allele folder in the current working directory will be used by default'
    )
    allele_concatenate_subparser.add_argument(
        '-c', '--concatenate_path',
        metavar='concatenate_path',
        default=os.path.join(os.getcwd(), 'concatenated_alleles'),
        help='Specify name and path of folder into which concatenated subunit files are to be placed. If not '
             'provided, the concatenated_alleles folder in the current working directory will be used'
    )
    allele_concatenate_subparser.set_defaults(func=allele_concatenate)
    # Get the arguments into an object
    arguments = setup_arguments(parser=parser)
    # Prevent the arguments being printed to the console (they are returned in order for the tests to work)
    sys.stderr = open(os.devnull, 'w', encoding='utf-8')
    return arguments


if __name__ == '__main__':
    cli()



#!/usr/bin/env python

"""
Unit and integration tests for allele_tools/stec.py Only testing the profile_reduce functionality
"""
import shutil
# Standard imports
from unittest.mock import patch
import argparse
import os

import pytest

# Local imports
from allele_tools.stec import \
    profile_reduce, \
    cli
from .test_0_profile_reduce import setup

assert vars(setup)['_pytestfixturefunction'].name == 'variables'


def test_missing_tilde(variables):
    from .test_0_profile_reduce import test_profile_reduce_missing_tilde_profile
    test_profile_reduce_missing_tilde_profile(variables=variables)


def test_missing_tilde_names(variables):
    from .test_0_profile_reduce import test_profile_reduce_missing_tilde_names
    test_profile_reduce_missing_tilde_names(variables=variables)


@patch('argparse.ArgumentParser.parse_args')
def test_stec_profile_reduce_integration(mock_args, variables):
    mock_args.return_value = argparse.Namespace(
        profile_file=variables.profile,
        gene_names=variables.names,
        output_folder=variables.output_path
    )
    arguments = cli()
    profile_reduce(args=arguments)
    assert os.path.isfile(variables.output_profiles)


def test_header(variables):
    from .test_0_profile_reduce import test_output_profile_header
    test_output_profile_header(variables=variables)


def test_sequence_type(variables):
    from .test_0_profile_reduce import test_output_profile_sequence_type
    test_output_profile_sequence_type(variables=variables)


def test_notes_header(variables):
    from .test_0_profile_reduce import test_output_notes_header
    test_output_notes_header(variables=variables)


def test_notes_novel(variables):
    from .test_0_profile_reduce import test_output_notes_novel
    test_output_notes_novel(variables=variables)


def test_notes_duplicate(variables):
    from .test_0_profile_reduce import test_output_notes_duplicate
    test_output_notes_duplicate(variables=variables)


def test_notes_length(variables):
    from .test_0_profile_reduce import test_output_notes_length
    test_output_notes_length(variables=variables)


@patch('argparse.ArgumentParser.parse_args')
def test_stec_profile_reduce_integration_no_genes(mock_args, variables):
    with pytest.raises(SystemExit):
        mock_args.return_value = argparse.Namespace(
            profile_file=variables.profile,
            gene_names=variables.names + '.fake',
            output_folder=variables.output_path
        )
        arguments = cli()
        profile_reduce(args=arguments)


@patch('argparse.ArgumentParser.parse_args')
def test_stec_profile_reduce_integration_no_folder(mock_args, variables):
    variables.fake_dir = os.path.join(variables.file_path, 'fake')
    fake_names = os.path.join(variables.fake_dir, variables.names)
    os.makedirs(variables.fake_dir)
    mock_args.return_value = argparse.Namespace(
        profile_file=variables.profile,
        gene_names=fake_names,
        output_folder=variables.output_path
    )
    arguments = cli()
    profile_reduce(args=arguments)
    assert os.path.isdir(variables.fake_dir)


@patch('argparse.ArgumentParser.parse_args')
def test_stec_profile_reduce_integration_no_file_or_folder(mock_args, variables):
    with pytest.raises(SystemExit):
        variables.fake_dir_fake = os.path.join(variables.file_path, 'fake_fake')
        fake_names = os.path.join(variables.fake_dir_fake, 'fake.txt')
        mock_args.return_value = argparse.Namespace(
            profile_file=variables.profile,
            gene_names=fake_names,
            output_folder=variables.output_path
        )
        arguments = cli()
        profile_reduce(args=arguments)


@patch('argparse.ArgumentParser.parse_args')
def test_stec_profile_reduce_integration_empty_file(mock_args, variables):
    with pytest.raises(SystemExit):
        variables.fake_dir_two = os.path.join(variables.file_path, 'fake_path')
        os.makedirs(variables.fake_dir_two)
        fake_names = os.path.join(variables.fake_dir_two, 'fake.txt')
        with open(fake_names, 'w', encoding='utf-8') as fake:
            fake.write('')
        mock_args.return_value = argparse.Namespace(
            profile_file=variables.profile,
            gene_names=fake_names,
            output_folder=variables.output_path
        )
        arguments = cli()
        profile_reduce(args=arguments)


def test_clean_up(variables):
    from .test_0_profile_reduce import test_output_clean_up
    test_output_clean_up(variables=variables)
    shutil.rmtree(variables.fake_dir)
    shutil.rmtree(variables.fake_dir_two)
    os.remove(variables.names + '.fake')
